from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union

from assonant.naming_standards import AcquisitionType


class IAssonantDataRetriever(ABC):
    """Assonant Data Retriever Interface.

    This interface defines what must be implemented by a class to be a Data Retriever
    in Assonant environment.
    """

    @abstractmethod
    def get_pv_data_by_acquisition_type(self, acquisition_type: AcquisitionType) -> Dict[str, Union[Any, List[Any]]]:
        """Return collected PV data related to specified AcquisitionType.

        The returned value is a dictionary as follow:

        {
            PV_NAME_1: [PV_VALUE_1, PV_VALUE_2, ...],
            PV_NAME_2: [PV_VALUE_1],
            ...
        }

        PS: PV values are always returned within a list, even if it is a single value.

        Args:
            acquisition_type (AcquisitionType): Target Acquisition Type used to select
            data from which Acquisition Type will be retrieved.

        Returns:
            Dict[str, Union[Any, List[Any]]]: Dictionary containing retrieved PV data.
        """



"""Assonant Undulator component data class."""

from .component import Component


class Undulator(Component):
    """Data class to handle all data required to define a undulator insertion device."""



from enum import Enum


class ValuePlaceholders(Enum):
    """Placeholder values to identify acquisition status or failure."""

    PV_NOT_CONNECTED = "PV not Connected"
    VALUE_NOT_SET = "Value not set"



"""Factory class to deal with NeXus Object Creation."""

from collections.abc import Iterable
from typing import Dict, List, Union

import numpy as np
from nexusformat.nexus import *

from assonant.data_classes import AssonantDataClass
from assonant.data_classes.components import (
    BVS,
    Aperture,
    Attenuator,
    Beam,
    Beamline,
    BeamStopper,
    BendingMagnet,
    Collimator,
    Component,
    Cryojet,
    Detector,
    DetectorChannel,
    DetectorModule,
    DetectorROI,
    Dewar,
    Entry,
    FresnelZonePlate,
    GraniteBase,
    Grating,
    Mirror,
    Monochromator,
    MonochromatorCrystal,
    MonochromatorVelocitySelector,
    Pinhole,
    Sample,
    Sensor,
    Shutter,
    Slit,
    StorageRing,
    Undulator,
    Wiggler,
)
from assonant.data_classes.data_handlers import Axis, DataField, DataHandler, TimeSeries

from .exceptions import NeXusObjectFactoryError


class NeXusObjectFactory:
    """NeXus Object Factory Class.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating NeXus object out of AssonantDataClass objects.
    """

    def create_nxobject(
        self, data_obj: AssonantDataClass, pack_into_nxroot: bool = False, convert_missing_data: bool = True
    ) -> NXobject:
        """Creates the respective Nexus object based on the passed AssonantDataClass specific type.

        Args:
            data_obj (AssonantDataClass): Data object which will be used for the
            NeXus object creation
            pack_into_nxroot (bool, optional): If returned object should be or not
            packed inside a NXroot. Defaults to False.
            convert_missing_data (bool, optional): If None values within DataField classes
            should be transformed or not in a np.nan value. Defaults to True.

        Raises:
            NeXusObjectFactoryError: The 'obj' argument type does not fit any of the
            supported types

        Returns:
            NXobject: NeXus object respective to the data object passed as the
            data_obj argument
        """
        # Control properti for missing vakue conversion within class methods
        self.convert_missing_data = convert_missing_data
        # print(data_obj, "==================================")
        if isinstance(data_obj, Entry):
            # print("Constructor Called: ExperimentEntry")
            nxobject = self._create_entry(data_obj)
        elif isinstance(data_obj, Beamline):
            # print("Constructor Called: Beamline")
            nxobject = self._create_beamline(data_obj)
        elif isinstance(data_obj, Sample):
            # print("Constructor Called: Sample")
            nxobject = self._create_sample(data_obj)
        elif isinstance(data_obj, Detector):
            # print("Constructor Called: Detector")
            nxobject = self._create_detector(data_obj)
        elif isinstance(data_obj, Monochromator):
            # print("Constructor Called: Monochromator")
            nxobject = self._create_monochromator(data_obj)
        elif isinstance(data_obj, BVS):
            # print("Constructor Called: BVS")
            nxobject = self._create_bvs(data_obj)
        elif isinstance(data_obj, Mirror):
            # print("Constructor Called: Mirror")
            nxobject = self._create_mirror(data_obj)
        elif isinstance(data_obj, Slit):
            # print("Constructor Called: Slit")
            nxobject = self._create_slit(data_obj)
        elif isinstance(data_obj, Axis):
            nxobject = self._create_axis(data_obj)
        elif isinstance(data_obj, TimeSeries):
            # print("Constructor Called: TimeSeries")
            nxobject = self._create_time_series(data_obj)
        elif isinstance(data_obj, DataField):
            # print("Constructor Called: Datafield")
            nxobject = self._create_data_field(data_obj)
        elif isinstance(data_obj, DetectorROI):
            # print("Constructor Called: DetectorROI")
            nxobject = self._create_detector_roi(data_obj)
        elif isinstance(data_obj, DetectorModule):
            # print("Constructor Called: DetectorModule")
            nxobject = self._create_detector_module(data_obj)
        elif isinstance(data_obj, DetectorChannel):
            # print("Constructor Called: DetectorChannel")
            nxobject = self._create_detector_channel(data_obj)
        elif isinstance(data_obj, MonochromatorCrystal):
            # print("Constructor Called: MonochromatorCrystal")
            nxobject = self._create_monochromator_crystal(data_obj)
        elif isinstance(data_obj, MonochromatorVelocitySelector):
            # print("Constructor Called: MonochromatorVelocitySelector")
            nxobject = self._create_monochromator_velocity_selector(data_obj)
        elif isinstance(data_obj, StorageRing):
            # print("Constructor Called: StorageRing")
            nxobject = self._create_storage_ring(data_obj)
        elif isinstance(data_obj, BendingMagnet):
            # print("Constructor Called: BendingMagnet")
            nxobject = self._create_bending_magnet(data_obj)
        elif isinstance(data_obj, Undulator):
            # print("Constructor Called: Undulator")
            nxobject = self._create_undulator(data_obj)
        elif isinstance(data_obj, Wiggler):
            # print("Constructor Called: Wiggler")
            nxobject = self._create_wiggler(data_obj)
        elif isinstance(data_obj, BeamStopper):
            # print("Constructor Called: BeamStopper")
            nxobject = self._create_beam_stopper(data_obj)
        elif isinstance(data_obj, Beam):
            # print("Constructor Called: Beam")
            nxobject = self._create_beam(data_obj)
        elif isinstance(data_obj, Shutter):
            # print("Constructor Called: Shutter")
            nxobject = self._create_shutter(data_obj)
        elif isinstance(data_obj, Collimator):
            # print("Constructor Called: Collimator")
            nxobject = self._create_collimator(data_obj)
        elif isinstance(data_obj, Attenuator):
            # print("Constructor Called: Attenuator")
            nxobject = self._create_attenuator(data_obj)
        elif isinstance(data_obj, Pinhole):
            # print("Constructor Called: Pinhole")
            nxobject = self._create_pinhole(data_obj)
        elif isinstance(data_obj, FresnelZonePlate):
            # print("Constructor Called: FresnelZonePlate")
            nxobject = self._create_fresnel_zone_plate(data_obj)
        elif isinstance(data_obj, Sensor):
            # print("Constructor Called: Sesor")
            nxobject = self._create_sensor(data_obj)
        elif isinstance(data_obj, Dewar):
            # print("Constructor Called: Dewar")
            nxobject = self._create_dewar(data_obj)
        elif isinstance(data_obj, Cryojet):
            # print("Constructor Called: Cryojet")
            nxobject = self._create_cryojet(data_obj)
        elif isinstance(data_obj, GraniteBase):
            # print("Constructor Called: GraniteBase")
            nxobject = self._create_granite_base(data_obj)
        elif isinstance(data_obj, Grating):
            # print("Constructor Called: Grating")
            nxobject = self._create_grating(data_obj)
        elif isinstance(data_obj, Aperture):
            # print("Constructor Called: Aperture")
            nxobject = self._create_aperture(data_obj)
        else:
            raise NeXusObjectFactoryError(
                f"NeXus object factory doesn't have an constructor method to deal with objects of type: {type(data_obj)}"
            )

        # Check if there are subgroups to be created based on special fields
        if hasattr(data_obj, "fields"):
            # print("Constructor Called: 'fields' Dictionary")
            nxobject = self._create_fields_from_dict(nxobject, data_obj.fields)

        if hasattr(data_obj, "positions"):
            # print("Constructor Called: Transformations")
            nxobject = self._create_transformations(nxobject, data_obj.positions)

        if hasattr(data_obj, "subcomponents"):
            # print("Constructor Called: Subcomponents")
            nxobject = self._create_subcomponents(nxobject, data_obj.subcomponents)

        return NXroot(nxobject) if pack_into_nxroot is True else nxobject

    def _create_entry(self, entry: Entry) -> NXentry:
        """Private method to create the NeXus object respective to any Entry data class.

        Args:
            entry (Entry): Data object containing entry data

        Returns:
            NXentry: NeXus object respective to the Entry data object
        """
        nxobject = NXentry(name=entry.name)
        return nxobject

    def _create_sample(self, sample: Sample) -> NXsample:
        """Private method to create the NeXus object respective to the Sample data class.

        Args:
            sample (Sample): Data object containing sample data

        Returns:
            NXsample: NeXus object respective to the Sample data object
        """
        nxobject = NXsample(name=sample.name)
        return nxobject

    def _create_detector(self, detector: Detector) -> NXdetector:
        """Private method to create the NeXus object respective to the Detector data class.

        Args:
            detector (Detector): Data object containing detector data

        Returns:
            NXdetector: NeXus object respective to the Detector data object
        """
        nxobject = NXdetector(name=detector.name)

        return nxobject

    def _create_monochromator(self, monochromator: Monochromator) -> NXmonochromator:
        """Private method to create the NeXus object respective to the Monochromator data class.

        Args:
            monochromator (Monochromator): Data object containing monochromator
            data

        Returns:
            NXmonochromator: NeXus object respective to the Monochromator data
            object
        """
        nxobject = NXmonochromator(name=monochromator.name)

        return nxobject

    def _create_bvs(self, bvs: BVS) -> NXdetector:
        """Private method to create the NeXus object respective to the BVS data class.

        Args:
            bvs (BVS): Data object containing BVS data

        Returns:
            NXdetector: NeXus object respective to the BVS data object
        """
        nxobject = NXdetector(name=bvs.name)

        return nxobject

    def _create_mirror(self, mirror: Mirror) -> NXmirror:
        """Private method to create the NeXus object respective to the Mirror data class.

        Args:
            mirror (Mirror): Data object containing mirror data

        Returns:
            NXmirror: NeXus object respective to the Mirror data object
        """
        nxobject = NXmirror(name=mirror.name)

        return nxobject

    def _create_slit(self, slit: Slit) -> NXslit:
        """Private method to create the NeXus object respective to the Slit data class.

        Args:
            slit (Slit): Data object containing slit data

        Returns:
            NXslit: NeXus object respective to the Slit data object
        """
        nxobject = NXslit(name=slit.name)

        return nxobject

    def _create_time_series(self, time_series: TimeSeries) -> NXlog:
        """Private method to create the NeXus object respective to the TimeSeries data class.

        Args:
            time_series (TimeSeries): Data object containing time_series data

        Returns:
            NXlog: NeXus object respective to the TimeSeries data object
        """
        nxobject = NXlog(
            value=self.create_nxobject(time_series.value),
            time=self.create_nxobject(time_series.timestamps),
        )

        return nxobject

    def _create_axis(self, axis: Axis) -> Union[NXfield, NXlog]:
        """Private method to create the NeXus object respective to the Axis data class.

        Args:
            axis (Axis): Data object containing data related to an axis

        Returns:
            Union[NXfield, NXlog]: NeXus object respective to the Axis data object. The
            returned object may change if depending if Axis 'value' field is was collected
            as a DataField or a TimeSeries.
        """
        if isinstance(axis.value, DataField):
            nxobject = self._create_data_field(axis.value)
            nxobject.attrs["transformation_type"] = axis.transformation_type.value
        elif isinstance(axis.value, TimeSeries):
            nxobject = self._create_time_series(axis.value)
            nxobject.value.attrs["transformation_type"] = axis.transformation_type.value

        return nxobject

    def _create_data_field(self, data_field: DataField) -> NXfield:
        """Private method to create the NeXus object respective to the Field data class.

        Args:
            data_field (DataField): Data object containing a single field
            data with its respective unit

        Returns:
            NXfield: NeXus object respective to the Field data object
        """

        if self.convert_missing_data is True:
            # print("Converting missing values from DataField...")
            data_field = self._convert_none_values_from_data_field(data_field=data_field)
            # print("Missing values conversion finished.")

        attrs = data_field.extra_metadata

        if data_field.unit is not None:
            attrs["unit"] = data_field.unit

        nxobject = NXfield(value=data_field.value, attrs=attrs)

        return nxobject

    def _create_detector_roi(self, detector_roi: DetectorROI) -> NXgroup:
        """Private method to create the NeXus object respective to the DetectorROI data class.

        Args:
            detector_roi (DetectorROI): Data object containing detector region of interest (ROI) data.

        Returns:
            NXgroup: NeXus object respective to the DetectorROI data object.
        """
        # Dev note: Currently, these type of data are stored in a contributed definition called NXregion. Due
        # to that the nexusformat package don't provide support for it so the nxclass attribute must be
        # set manually.
        nxobject = NXgroup(name=detector_roi.name, nxclass="NXregion")
        return nxobject

    def _create_detector_module(self, detector_module: DetectorModule) -> NXdetector_module:
        """Private method to create the NeXus object respective to the DetectorModule data class.

        Args:
            detector_module (DetectorModule): Data object containing detector module data.

        Returns:
            NXdetector_module: NeXus object respective to the DetectorModule data object.
        """
        nxobject = NXdetector_module(name=detector_module.name)
        return nxobject

    def _create_detector_channel(self, detector_channel: DetectorChannel) -> NXdetector_channel:
        """Private method to create the NeXus object respective to the DetectorChannel data class.

        Args:
            detector_channel (DetectorChannel): Data object containing detector channel data.

        Returns:
            NXdetector_channel: NeXus object respective to the DetectorChannel data object.
        """
        nxobject = NXdetector_channel(name=detector_channel.name)
        return nxobject

    def _create_monochromator_crystal(self, monochromator_crystal: MonochromatorCrystal) -> NXcrystal:
        """Private method to create the NeXus object respective to the MonochromatorCrystal data class.

        Args:
            monochromator_crystal (MonochromatorCrystal):  Data object containing monochromator crystal data.

        Returns:
            NXcrystal: NeXus object respective to the MonochromatorCrystal data object.
        """
        nxobject = NXcrystal(name=monochromator_crystal.name)
        return nxobject

    def _create_monochromator_velocity_selector(
        self, monochromator_velocity_selector: MonochromatorVelocitySelector
    ) -> NXvelocity_selector:
        """Private method to create the NeXus object respective to the MonochromatorVelocitySelector data class.

        Args:
            monochromator_velocity_selector (MonochromatorVelocitySelector): Data object containing monochromator
            velocity selector data.

        Returns:
            NXvelocity_selector: NeXus object respective to the MonochromatorVelocitySelector data object.
        """
        nxobject = NXvelocity_selector(name=monochromator_velocity_selector.name)
        return nxobject

    def _create_storage_ring(self, storage_ring: StorageRing) -> NXsource:
        """Private method to create the NeXus object respective to the StorageRing data class.

        Args:
            storage_ring (StorageRing): Data object containing storage ring data.

        Returns:
            NXsource: NeXus object respective to the StorageRing data object.
        """
        nxobject = NXsource(name=storage_ring.name)
        return nxobject

    def _create_bending_magnet(self, bending_magnet: BendingMagnet) -> NXbending_magnet:
        """Private method to create the NeXus object respective to the BendingMagnet data class.

        Args:
            bending_magnet (BendingMagnet): Data object containing bending magnet data.

        Returns:
            NXbending_magnet: NeXus object respective to the BendingMagnet data object.
        """
        nxobject = NXbending_magnet(name=bending_magnet.name)
        return nxobject

    def _create_undulator(self, undulator: Undulator) -> NXinsertion_device:
        """Private method to create the NeXus object respective to the Undulator data class.

        Args:
            undulator (Undulator): Data object containing undulator data.

        Returns:
            NXinsertion_device: NeXus object respective to the Undulator data object.
        """
        nxobject = NXinsertion_device(name=undulator.name, type="undulator")
        return nxobject

    def _create_wiggler(self, wiggler: Wiggler) -> NXinsertion_device:
        """Private method to create the NeXus object respective to the Wiggler data class.

        Args:
            wiggler (Wiggler): Data object containing wiggler data.

        Returns:
            NXinsertion_device: NeXus object respective to the Wiggler data object.
        """
        nxobject = NXinsertion_device(name=wiggler.name, type="wiggler")
        return nxobject

    def _create_beam_stopper(self, beam_stopper: BeamStopper) -> NXbeam_stop:
        """Private method to create the NeXus object respective to the BeamStopper data class.

        Args:
            beam_stopper (BeamStopper): Data object containing beam_stopper data.

        Returns:
            NXbeam_stop: NeXus object respective to the BeamStopper data object.
        """
        nxobject = NXbeam_stop(name=beam_stopper.name)
        return nxobject

    def _create_beam(self, beam: Beam) -> NXbeam:
        """Private method to create the NeXus object respective to the Beam data class.

        Args:
            beam (Beam): Data object containing beam data.

        Returns:
            NXbeam: NeXus object respective to the Beam data object.
        """
        nxobject = NXbeam(name=beam.name)
        return nxobject

    def _create_shutter(self, shutter: Shutter) -> NXbeam_stop:
        """Private method to create the NeXus object respective to the Shutter data class.

        Args:
            shutter (Shutter): Data object containing shutter data.

        Returns:
            NXbeam_stop: NeXus object respective to the Shutter data object.
        """
        nxobject = NXbeam_stop(name=shutter.name)
        return nxobject

    def _create_collimator(self, collimator: Collimator) -> NXcollimator:
        """Private method to create the NeXus object respective to the Collimator data class.

        Args:
            collimator (Collimator): Data object containing collimator data.

        Returns:
            NXcollimator: NeXus object respective to the Collimator data object.
        """
        nxobject = NXcollimator(name=collimator.name)
        return nxobject

    def _create_attenuator(self, attenuator: Attenuator) -> NXattenuator:
        """Private method to create the NeXus object respective to the Attenuator data class.

        Args:
            attenuator (Attenuator): Data object containing attenuator data.

        Returns:
            NXattenuator: NeXus object respective to the Attenuator data object.
        """
        nxobject = NXattenuator(name=attenuator.name)
        return nxobject

    def _create_pinhole(self, pinhole: Pinhole) -> NXpinhole:
        """Private method to create the NeXus object respective to the Pinhole data class.

        Args:
            pinhole (Pinhole): Data object containing pinhole data.

        Returns:
            NXpinhole: NeXus object respective to the Pinhole data object.
        """
        nxobject = NXpinhole(name=pinhole.name)
        return nxobject

    def _create_fresnel_zone_plate(self, fresnel_zone_plate: FresnelZonePlate) -> NXfresnel_zone_plate:
        """Private method to create the NeXus object respective to the FresnelZonePlate data class.

        Args:
            fresnel_zone_plate (FresnelZonePlate): Data object containing fresnel_zone_plate data.

        Returns:
            NXfresnel_zone_plate: NeXus object respective to the FresnelZonePlate data object.
        """
        nxobject = NXfresnel_zone_plate(name=fresnel_zone_plate.name)
        return nxobject

    def _create_sensor(self, sensor: Sensor) -> NXsensor:
        """Private method to create the NeXus object respective to the Sensor data class.

        Args:
            sensor (Sensor): Data object containing sensor data.

        Returns:
            NXsensor: NeXus object respective to the Sensor data object.
        """
        nxobject = NXsensor(name=sensor.name)
        return nxobject

    def _create_grating(self, grating: Grating) -> NXgrating:
        """Private method to create the NeXus object respective to the Grating data class.

        Args:
            grating (Grating): Data object containing grating data.

        Returns:
            NXgrating: NeXus object respective to the Grating data object.
        """
        nxobject = NXgrating(name=grating.name)
        return nxobject

    def _create_aperture(self, aperture: Aperture) -> NXaperture:
        """Private method to create the NeXus object respective to the Aperture data class.

        Args:
            aperture (Aperture): Data object containing aperture data.

        Returns:
            NXaperture: NeXus object respective to the Aperture data object.
        """
        nxobject = NXaperture(name=aperture.name)
        return nxobject

    def _create_dewar(self, dewar: Dewar) -> NXgroup:
        """Private method to create the NeXus object respective to the Dewar data class.

        Args:
            dewar (Dewar): Data object containing dewar data.

        Returns:
            NXgroup: NeXus object respective to the Dewar data object.
        """
        # Dev note: Currently, there is no base or contributed definition on NeXus for this type of component.
        # Due to that, a temporary nxclass value will be set to allow future automatized replacement if a
        # standard for it is defined.
        nxobject = NXgroup(name=dewar.name, nxclass="NXcustom_dewar")
        return nxobject

    def _create_cryojet(self, cryojet: Cryojet) -> NXgroup:
        """Private method to create the NeXus object respective to the Cryojet data class.

        Args:
            cryojet (Cryojet): Data object containing dewar data.

        Returns:
            NXgroup: NeXus object respective to the Cryojet data object.
        """
        # Dev note: Currently, there is no base or contributed definition on NeXus for this type of component.
        # Due to that, a temporary nxclass value will be set to allow future automatized replacement if a
        # standard for it is defined.
        nxobject = NXgroup(name=cryojet.name, nxclass="NXcustom_cryojet")
        return nxobject

    def _create_granite_base(self, granite_base: GraniteBase) -> NXgroup:
        """Private method to create the NeXus object respective to the Cryojet data class.

        Args:
            granite_base (GraniteBase): Data object containing granite base data.

        Returns:
            NXgroup: NeXus object respective to the GraniteBase data object.
        """
        # Dev note: Currently, there is no base or contributed definition on NeXus for this type of component.
        # Due to that, a temporary nxclass value will be set to allow future automatized replacement if a
        # standard for it is defined.
        nxobject = NXgroup(name=granite_base.name, nxclass="NXcustom_granite_base")
        return nxobject

    def _create_beamline(self, beamline: Beamline) -> NXinstrument:
        """Private method to create the NeXus object respective to the Beamline data class.

        Args:
            beamline (Beamline): Data object containing beamline data.

        Returns:
            NXinstrument: NeXus object respective to the Beamline data object.
        """
        nxobject = NXinstrument(name="instrument")
        nxobject["name"] = beamline.name
        return nxobject

    def _create_transformations(self, nxobject: NXobject, positions: Dict[str, Axis]) -> NXobject:
        """Create a NXtransformations group to store positioning data passed as a list of Axis objects.

        Args:
            nxobject (NXobject): NeXus object containing current objects
            positions (dict[str, Axis]): Dict of Axis objects containing each of them the positioning data related to a
            specific monitored axis.

        Returns:
           NXobject: Respective NXobject with a NXcollection group containing all fields passed.
        """
        if positions != {}:
            nxtransformations = NXtransformations(name="transformations")
            for key in positions:
                try:
                    nxtransformations[key] = self.create_nxobject(positions[key])
                except Exception:
                    raise NeXusObjectFactoryError(
                        f"Error ocurred when creating NXobject for {key} Axis from {nxobject._name} object"
                    )
            nxobject.insert(nxtransformations)
        return nxobject

    def _create_fields_from_dict(self, nxobject: NXobject, data_dict: Dict[str, DataHandler]) -> NXobject:
        """Create fields to store data passed on 'fields' dictionary from object.

        Each key, value tuple from the dictionary will be converted in a field with name equal to the tuple key.

        Args:
            nxobject (NXobject): NeXus object containing current objects
            data_dict (Dict): Dictionary containing data that will be converted to group fields.

        Raises:
            NeXusObjectFactoryError: Happens when a not supported AssonantDataClass is used on "data" dict values.

        Returns:
            NXobject: NXobject passed as input with fields set in it based on passed dictionary.
        """
        if data_dict != {}:
            for key in data_dict.keys():
                if isinstance(data_dict[key], DataField):
                    nxobject[key] = self.create_nxobject(data_dict[key])
                elif isinstance(data_dict[key], TimeSeries):
                    nxobject[key] = self.create_nxobject(data_dict[key])
                else:
                    raise NeXusObjectFactoryError(f"Expected 'data_dict[{key}]' to be a DataField or a TimeSeries")
        return nxobject

    def _create_subcomponents(self, nxobject: NXobject, subcomponents: Dict[str, Component]) -> NXobject:
        """Private method to create NeXus objects relative to subcomponents from passed NXobject.

        Args:
            subcomponents (Dict[str, Component]): Dict containing Component object which will be
            converted to their respective NXobjects and inserted on passed NXobject.

        Returns:
            NXobject: Passed NXobject updated with subcomponents groups created into it.
        """
        if subcomponents != []:
            for subcomponent_name in subcomponents:
                nxobject.insert(self.create_nxobject(subcomponents[subcomponent_name]))
        return nxobject

    def _convert_none_values_from_list(self, lst: List) -> List:
        """Transform None value within a list into np.nan.

        Args:
            lst (List): List containing values to be verified.

        Raises:
            NeXusObjectFactoryError: Raised when any error occurs when iterating the list.

        Returns:
            List: List with None values transformed into np.nan.
        """
        try:
            return [value if value is not None else np.nan for value in lst]
        except Exception as e:
            raise NeXusObjectFactoryError(f"An error ocurred when iterating list of{type(lst)} type.") from e

    def _convert_none_values_from_dict(self, d: Dict) -> Dict:
        """Transform None value within a dictionary into np.nan.

        Args:
            d (Dict): Dictionary containing values to be verified.

        Raises:
            NeXusObjectFactoryError: Raised when any error occurs when iterating dictionary elements.

        Returns:
            Dict: Dictionary with None values transformed into np.nan.
        """
        try:
            for key in d:
                if isinstance(d[key], Iterable):
                    d[key] = self._convert_none_values_from_iterable(iterable=d[key])
                elif d[key] is None:
                    d[key] = np.nan
            return d
        except Exception as e:
            raise NeXusObjectFactoryError("An error ocurred when iterating dict.") from e

    def _convert_none_values_from_iterable(self, iterable: Iterable) -> Iterable:
        if isinstance(iterable, str):
            pass  # No changed needed for this cases
        elif isinstance(iterable, List) or isinstance(iterable, np.ndarray):
            iterable = self._convert_none_values_from_list(iterable)
        elif isinstance(iterable, Dict):
            iterable = self._convert_none_values_from_dict(iterable)
        else:
            raise NeXusObjectFactoryError(
                f"Iterable from type {type(iterable)} is not suported for None values convertion."
            )

        return iterable

    def _convert_none_values_from_data_field(self, data_field: DataField) -> DataField:
        """Transform None values within a DataField value field into np.nan.

        Args:
            data_field (DataField): DataField object which value field will be verified.

        Returns:
            DataField: DataField object with None values within its value field transformed into np.nan.
        """
        try:
            if isinstance(data_field.value, Iterable):
                data_field.value = self._convert_none_values_from_iterable(iterable=data_field.value)
            elif data_field.value is None:
                data_field.value = np.nan
        except Exception as e:
            raise NeXusObjectFactoryError("An error occured when converting DataField None values.") from e

        if data_field.unit is None:
            data_field.unit = np.nan

        if data_field.extra_metadata is None:
            data_field.extra_metadata = {}

        return data_field



"""Expermiental Technique names definitions."""

from enum import Enum


class ExperimentalTechniquesAcronyms(Enum):
    """Experimental Technique acronyms standards."""

    ARPES = "ARPES"
    CD = "CD"
    CT = "CT"
    MX = "MX"
    PDF = "PDF"
    PEEM = "PEEM"
    RIXS = "RIXS"
    SAXS = "SAXS"
    STMX = "STMX"
    USAXS = "USAXS"
    WAXS = "WAXS"
    XAS = "XAS"
    XES = "XES"
    XEOL = "XEOL"
    XMCD = "XMCD"
    XPS = "XPS"
    XPCS = "XPCS"
    XRD = "XRD"
    XRF = "XRF"


class ExperimentalTechniquesFullNames(Enum):
    """Experimental Technique full name standards."""

    ARPES = "Angle-Resolved Photoemission Spectroscopy"
    CD = "Circular Dichroism Spectroscopy"
    CT = "Computed Tomography"
    MX = "Macromolecular Crystallography"
    PDF = "Pair Distribution Function"
    PEEM = "Photoemission Electron Microscopy"
    RIXS = "Resonant Inelastic X-ray Scattering"
    SAXS = "Small Angle X-ray Scattering"
    STMX = "Scanning Transmission X-ray Microscopy"
    USAXS = "Ultra-Small Angle X-ray Scattering"
    WAXS = "Wide-Angle X-ray Scattering"
    XAS = "X-ray Absortion Spectroscopy"
    XES = "X-ray Emission Spectroscopy"
    XEOL = "X-ray Excited Optical Luminescence"
    XMCD = "X-ray Magnetic Circular Dichroism"
    XPS = "X-ray Photonelectron Spectroscopy"
    XPCS = "X-ray Photon Correlation Spectroscopy"
    XRD = "X-ray Diffraction"
    XRF = "X-ray Fluorescence"



from abc import ABC
from typing import Callable, Dict, List

from ._nexus_file_writer import NexusFileWriter
from .assonant_file_writer_interface import IAssonantFileWriter
from .exceptions import FileWriterFactoryError


class FileWriterFactory(ABC):
    """Abstract Factory for creating Assonant File Writers.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating Assonant Component objects.

    Uses a registry-based approach so that new file formats can be added
    without modifying the core creation logic (Open/Closed principle).

    Example:
        FileWriterFactory.register_file_format("myformat", MyFileWriter)
        writer = FileWriterFactory.create_file_writer("myformat")
    """

    _creators: Dict[str, Callable[[], IAssonantFileWriter]] = {}

    @staticmethod
    def register_file_format(file_format: str, creator: Callable[[], IAssonantFileWriter]) -> None:
        """Register a file format with its creator callable.

        Args:
            file_format (str): File format key (lowercase recommended).
            creator (Callable[[], IAssonantFileWriter]): Function or class returning a
            IAssonantFileWriter instance.
        """
        if not callable(creator):
            raise FileWriterFactoryError("Creator must be callable and return an IAssonantFileWriter instance.")

        FileWriterFactory._creators[file_format] = creator

    @staticmethod
    def create_file_writer(file_format: str) -> IAssonantFileWriter:
        """Create a file writer instance for the given format.

        Args:
            file_format (str): Target file format.

        Raises:
            FileWriterFactoryError: If the format is unsupported.

        Returns:
            IAssonantFileWriter: File writer instance.
        """
        try:
            creator = FileWriterFactory._creators[file_format]
        except KeyError:
            raise FileWriterFactoryError(
                f"'{file_format}' file format isn't currently supported. "
                f"Supported formats: {list(FileWriterFactory._creators.keys())}"
            )

        return creator()

    @staticmethod
    def get_supported_file_formats() -> List[str]:
        """Return a list of supported file formats."""
        return list(FileWriterFactory._creators.keys())


# === Register built-in formats here ===
FileWriterFactory.register_file_format("nexus", NexusFileWriter)



"""Assonant Hierarchizer Exceptions.

This submodule defines Exceptions to signalized errors wthin the Hierarchizer submodule.
"""

from .hierarchizer_exceptions import AssonantHierarchizerError

__all__ = ["AssonantHierarchizerError"]



"""Tests focused on validating Assonant PathBuilder methods"""

from pathlib import Path

import pytest

from assonant.naming_standards import BeamlineName
from assonant.path_builder import BeamlinePathBuilder


@pytest.fixture(scope="function")
def example_path_builder() -> BeamlinePathBuilder:
    """
    Provides a BeamlinePathBuilder to be used on tests.
    """
    return BeamlinePathBuilder()


@pytest.fixture(scope="session")
def example_proposal_id() -> str:
    """
    Provides a standard example proposal ID to be used on tests.
    """
    return "20250826"


@pytest.fixture(scope="session")
def example_beamline_name() -> str:
    """
    Provides a standard example beamline name to be used on tests.
    """
    return BeamlineName.CARNAUBA.value.lower()


@pytest.fixture(scope="session")
def expected_data_schema_path(example_beamline_name) -> str:
    """
    Provides the expected data schema path considering the example beamline name.
    """
    return str(Path(f"/ibira/lnls/beamlines/{example_beamline_name}/apps/assonant/metadata.csv"))


@pytest.fixture(scope="session")
def expected_proposal_path(example_beamline_name, example_proposal_id) -> str:
    """
    Provides the expected proposal path considering the example beamline name and example proposal id.
    """
    return str(Path(f"/ibira/lnls/beamlines/{example_beamline_name}/proposals/{example_proposal_id}"))


def test_set_beamline_name(example_path_builder, example_beamline_name):
    """
    Validate that BeamlinePathBuilder set_beamline_name method correctly set
    beamline_name property.

    Expected Behavior:
        - BeamlinePathBuilder beamline_name property must have the same value
        than the passed example_beamline_name
    """
    example_path_builder.set_beamline_name(example_beamline_name)
    assert example_path_builder.beamline_name == example_beamline_name


def test_set_proposal_id(example_path_builder, example_proposal_id):
    """
    Validate that BeamlinePathBuilder set_proposal method correctly set
    proposal_id property.

    Expected Behavior:
        - BeamlinePathBuilder proposal_id property must have the same value
        than the passed example_proposal_id
    """
    example_path_builder.set_proposal(example_proposal_id)
    assert example_path_builder.proposal_id == example_proposal_id


def test_build_path_to_data_schema_success(example_path_builder, example_beamline_name, expected_data_schema_path):
    """
    Validate that BeamlinePathBuilder build_path_to_data_schema method
    correctly generate path to beamline data_schema file

    Expected Behavior:
        - BeamlinePathBuilder generated path to data_schema file must
        be the same as the expected data schema path.
    """
    example_path_builder.set_beamline_name(example_beamline_name)
    assert example_path_builder.build_path_to_data_schema() == expected_data_schema_path


def test_build_path_to_data_schema_without_beamline(example_path_builder):
    """
    Validate that BeamlinePathBuilder correctly asserts and raises exception
    if build_path_to_data_schema method is called before setting target
    beamline_name property.

    Expected Behavior:
        - BeamlinePathBuilder must raise an AssertionError signalizing that the
        beamline_name property was not set.
    """
    with pytest.raises(AssertionError, match="Beamline name property must be set"):
        example_path_builder.build_path_to_data_schema()


def test_build_path_to_proposal_success(
    example_path_builder, example_beamline_name, example_proposal_id, expected_proposal_path
):
    """
    Validate that BeamlinePathBuilder build_path_to_proposal method
    correctly generate path to proposal directory

    Expected Behavior:
        - BeamlinePathBuilder generated path to proposal directory file must
        be the same as the expected proposal path.
    """
    example_path_builder.set_beamline_name(example_beamline_name)
    example_path_builder.set_proposal(example_proposal_id)
    assert example_path_builder.build_path_to_proposal() == expected_proposal_path


def test_build_path_to_proposal_without_proposal(example_path_builder, example_beamline_name):
    """
    Validate that BeamlinePathBuilder correctly asserts and raises exception
    if build_path_to_proposal method is called before setting target
    proposal property.

    Expected Behavior:
        - BeamlinePathBuilder must raise an AssertionError signalizing that the
        proposal_id property was not set.
    """

    example_path_builder.set_beamline_name(example_beamline_name)
    with pytest.raises(AssertionError, match="Proposal ID property must be set"):
        example_path_builder.build_path_to_proposal()


def test_build_path_to_proposal_without_beamline(example_path_builder, example_proposal_id):
    """
    Validate that BeamlinePathBuilder correctly asserts and raises exception
    if build_path_to_proposal method is called before setting target
    beamline_name property.

    Expected Behavior:
        - BeamlinePathBuilder must raise an AssertionError signalizing that the
        beamline_name property was not set.
    """

    example_path_builder.set_proposal(example_proposal_id)
    with pytest.raises(AssertionError, match="Beamline name property must be set"):
        example_path_builder.build_path_to_proposal()


def test_reset_functionality(example_path_builder, example_beamline_name, example_proposal_id):
    """
    Validate that BeamlinePathBuilder reset method correctly resets its properties.

    Expected Behavior:
        - BeamlinePathBuilder properties must be reset to None value after
        reset method is called.
    """
    assert example_path_builder.beamline_name is None
    assert example_path_builder.proposal_id is None
    example_path_builder.set_beamline_name(example_beamline_name)
    example_path_builder.set_proposal(example_proposal_id)
    assert example_path_builder.beamline_name is not None
    assert example_path_builder.proposal_id is not None
    example_path_builder.reset()
    assert example_path_builder.beamline_name is None
    assert example_path_builder.proposal_id is None



"""Assonant DataField class."""

from typing import Dict, List, Optional, Type, Union

import numpy as np

from ..exceptions import AssonantDataClassesError
from .data_handler import DataHandler


class DataField(DataHandler):
    """Data class to handle any type of base data, such as integers, floats, numpy arrays and lists."""

    value: Union[int, float, str, List, Type[np.ndarray], None]
    unit: Optional[str] = None
    extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {}

    def set_value(self, value: Union[int, float, str, List, Type[np.ndarray], None]):
        """Set value to 'value' property from DataField.

        Args:
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value to be set to
            DataField 'value' property.
        """
        self.value = value

    def set_unit(self, value: str):
        """Set value to 'unit' property from DataField.

        Args:
            value (str): Value to be set to DataField 'unit' property
        """
        self.unit = value

    def add_extra_metadata(self, metadata_name: str, value: Union[int, float, str, List, Type[np.ndarray], None]):
        """Add new metadata with passed name to extra_metadata dict from DataField.

        Args:
            metadata_name (str): Name of passed metadata to be created.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value to be set to new extra_metadata.

        Raises:
            AssonantDataClassesError: Raised when a passed metadata name conflicts with an already existing key from
            'extra_metadata' dictionary from DataField.
        """

        if metadata_name not in self.extra_metadata.keys():
            self.extra_metadata[metadata_name] = value
        else:
            raise AssonantDataClassesError(
                f"'{metadata_name}' metadata already exists on DataFielf extra_metadata dict!"
            )

    def set_extra_metadata(self, metadata_name: str, value: Union[int, float, str, List, Type[np.ndarray], None]):
        """Set value to passed extra_metadata with passed name within DataField extra_metadata dict.

        Args:
            metadata_name (str): Name of existing metadata that value will be set.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value to be set to existing extra_metadata.

        Raises:
            AssonantDataClassesError: Raised when a passed metadata name doesn't exists in 'extra_metadata'
            dictionary from DataField.
        """
        if metadata_name in self.extra_metadata.keys():
            self.extra_metadata[metadata_name] = value
        else:
            raise AssonantDataClassesError(f"'{metadata_name}' metadata don't exists on DataField extra_metadata dict!")



import os

from nexusformat.nexus import NXentry, nxopen

from assonant._nexus import NeXusObjectFactory
from assonant.data_classes import AssonantDataClass

from .assonant_file_writer_interface import IAssonantFileWriter
from .exceptions import (
    AssonantFileWriterError,
    FileAlreadyExistsError,
    NeXusObjectAlreadyExistsError,
)


class NexusFileWriter(IAssonantFileWriter):
    """NeXus File Writer.

    File Writer specialized on accessing data from AssonantDataClasses and
    write into a HDF5 file following NeXus data model standards.
    """

    def __init__(self):
        self.nexus_factory = NeXusObjectFactory()
        self.file_extension = ".nxs"

    def _try_nxobject_insertion(self, parent, nxobject):
        """
        Attempts to insert an NXobject into a parent NXobject, handling conflicts
        based on the type of the NXobject.

        If the NXobject is not an NXentry and a conflict exists (i.e., an object
        with the same name already exists in the parent), an error is raised. If
        the NXobject is an NXentry and a conflict exists, the method recursively
        attempts to insert the components of the source NXentry into the target
        NXentry.

        Args:
            parent (NXobject): The parent NXobject where the nxobject is to be inserted.
            nxobject (NXobject): The NXobject to be inserted into the parent.

        Raises:
            NeXusObjectAlreadyExistsError: If a non-NXentry object with the same name
                already exists in the parent.
        """
        if nxobject.nxname in parent and not isinstance(nxobject, NXentry):
            # If Non-Entry object conflicts, raise error instantly
            raise NeXusObjectAlreadyExistsError(
                f"Non Entry Object '{nxobject.nxname}' already exists in parent '{parent.nxname}'"
            )
        elif nxobject.nxname in parent and isinstance(nxobject, NXentry):
            # If Entry object conflicts, try to insert components within source Entry
            # to target Entry recursively.
            for nx_obj in nxobject._entries.values():
                self._try_nxobject_insertion(parent[nxobject.nxname], nx_obj)
        else:
            # If no conflict, insert object directly
            parent.insert(nxobject)

    def write_data(self, filepath: str, filename: str, data: AssonantDataClass):
        """Writes data into a NeXus file.

        This method creates a new NeXus file if it does not exist, or appends data
        to the NXroot object of an existing file. The data to be written is provided
        as an instance of `AssonantDataClass`.

            filepath (str): The directory path where the NeXus file will be saved.
            filename (str): The name of the NeXus file (without extension).
            data (AssonantDataClass): An instance of `AssonantDataClass` containing
                the data to be written to the file.

        Raises:
            FileAlreadyExistsError: Raised if there is a conflict when attempting
                to write a NeXus object, indicating that the file already exists.
                File is considered to exist if an ExperimentStage Entry and its
                respective subentry conflicts.
            AssonantFileWriterError: Raised for any other errors encountered while
                writing data to the file.
        """
        nexus_filepath = os.path.join(filepath, os.path.splitext(filename)[0]) + self.file_extension
        mode = "rw" if os.path.isfile(nexus_filepath) else "a"

        try:
            with nxopen(nexus_filepath, mode=mode) as nxf:
                nxobject = self.nexus_factory.create_nxobject(data)
                # Try to insert Entry related to experiment stage into file
                self._try_nxobject_insertion(nxf, nxobject)
        except NeXusObjectAlreadyExistsError as e:
            raise FileAlreadyExistsError(
                f"Due to conflict on writing NeXus object, file {nexus_filepath} is considered to be already existent!"
            ) from e
        except Exception as e:
            raise AssonantFileWriterError(f"NeXusFileWriter failed to write data into {filepath}") from e

    def get_file_extension(self) -> str:
        """Getter method for the file extension of currently handled file writer.

        Returns:
            str: File extenstion of urrently handled file writer.
        """
        return self.file_extension



"""Assonant DataHandler abstract class."""

from typing import Any

from ..assonant_data_class import AssonantDataClass


# TODO: Make this class abstract
class DataHandler(AssonantDataClass):
    """Abstract class that creates the base common requirements to define an Assonant DataHandler."""

    value: Any

    def set_value(self, value: Any):
        """Set value to DataHandler 'value' property.

        This MUST be overrided by child classes when specific logic based on the
        DataHandler type need to be performed.

        Args:
            value (Any): Value to be set to the 'value' property from the DataHandler
        """
        self.value = value



"""Assonant Axis data handler."""

from typing import List, Type, Union

import numpy as np

from ..enums import TransformationType
from .data_field import DataField
from .data_handler import DataHandler
from .time_series import TimeSeries


class Axis(DataHandler):
    """Data class to handle data related to an axis position.

    When axis position is static the 'value' field must contain a single value and when it varies over time, it
    should be an array or TimeSeries.
    """

    transformation_type: TransformationType
    value: Union[DataField, TimeSeries]

    def set_value(self, value: Union[int, float, str, List, Type[np.ndarray], None]):
        """Set value to 'value' property from Axis.

        Args:
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value to bet set to
            Axis 'value' property.
        """
        self.value.set_value(value)

    def set_transformation_type(self, value: TransformationType):
        """Set value to 'transformation_type' property from Axis.

        Args:
            value (TransformationType): Enum of value respective to Axis transformation type.
        """
        self.transformation_type = value



from enum import Enum


class ValuePlaceholders(Enum):
    """Placeholder values to identify acquisition status or failure."""

    PV_NOT_CONNECTED = "PV not Connected"  # Used when PV was disconnected during acquisition tentative
    VALUE_NOT_SET = "Value not set"  # Used as placeholder value when a value was not set to a field yet



"""Factories used to abstract and centralize Assonant data classes instanciation process."""

import importlib
from typing import Dict, List, Optional, Type, Union

import numpy as np

from .components import Component
from .data_handlers import Axis, DataField, TimeSeries
from .enums import TransformationType
from .exceptions import AssonantComponentFactoryError, AssonantDataHandlerFactoryError


class AssonantComponentFactory:
    """Assonant Component Factory Class.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating Assonant Component objects.
    """

    @staticmethod
    def create_component_by_class_name(class_name: str, component_name: str) -> Component:
        """Create an Assonant data class Component based on the class name passed in a string.

        Args:
            class_name (str): String with value equal to the name of a AssonantDataClass type.
            component_name (str): Name given to the component that will be created.

        Raises:
            AssonantComponentFactoryException: Passed 'class_name' argument is invalid.

        Returns:
            Component: Named Component object of type equals to the 'class_name' parameter.
        """
        try:
            component_class = getattr(importlib.import_module("assonant.data_classes.components"), class_name)
            component = component_class(name=component_name)
        except Exception as e:
            raise AssonantComponentFactoryError(
                f"An Error ocurred when creating Assonant Component by the following class name: {class_name}"
            ) from e
        return component

    @staticmethod
    def create_component_by_class_type(class_type: Component, component_name: str) -> Component:
        """Create an Assonant data class Component based on the class type passed.

        Args:
            class_type (Component): Assonant Component type of the class that will be created.
            component_name (str): Name given to the component that will be created.

        Raises:
            AssonantComponentFactoryException: Passed 'class_type' argument is invalid.

        Returns:
            Component: Named Component object of 'class_type' type.
        """
        try:
            component = class_type(name=component_name)
        except Exception as e:
            raise AssonantComponentFactoryError(
                f"An Error ocurred when creating Assonant Component by the following class type: {class_type}"
            ) from e
        return component


class AssonantDataHandlerFactory:
    """NeXus Data Handler Factory Class.

    Class that implements the factory design pattern
    (https://refactoring.guru/design-patterns/factory-method) to fully abstract
    the procedure of creating Assonant Data Handler objects.
    """

    @staticmethod
    def create_data_handler(
        value: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        transformation_type: Optional[TransformationType] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
        timestamps: Union[int, float, str, List, Type[np.ndarray], None] = None,
        timestamps_unit: Optional[str] = None,
        timestamp_extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ) -> Union[DataField, Axis, TimeSeries]:
        """Create the best fitting DataHandler object based on passed data.

        Args:
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to collected data.
            unit (Optional[str], optional): Measurement unit related to value parameter. Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]], optional): Dict
            containing any aditional metadata related to collected data. Defaults to {}.
            transformation_type (Optional[TransformationType], optional): TransformationType related to Axis data.
            Defaults to None.
            timestamps (Union[int, float, str, List, Type[np.ndarray], None], optional): Timestamps related to
            collected data. Defaults to None.
            timestamps_unit (Optional[str], optional): Measurement unit related to timestamps parameter.
            Defaults to None.
            timestamp_extra_metadata (Dict[str, Union[int, float, str, List, Type[np.ndarray], None]], optional):
            Dictionary containing extra metadata about timestamps field. Defaults to {}.

        Returns:
            Union[DataField, Axis, TimeSeries]: DataHandler object depeding on passed parameters:
            1) If timestamps != None AND transformationType != None method returns a TimeSeries with 'value'
            being an Axis.
            2) If timestamps != None AND transformationType == None method returns a  TimeSeries with 'value'
            being a DataField.
            3) If timestamps == None AND transformationType != None method returns an Axis.
            4) If timestamps == None AND transformationType == None method returns a DataField.

        """
        # Check key optional fields that changes the used DataHandler class
        if timestamps is None and timestamps_unit is None and timestamp_extra_metadata == {}:
            # It is not a TimeSeries monitored ...
            if transformation_type is not None:
                # Axis
                new_data_handler = AssonantDataHandlerFactory.create_axis(
                    transformation_type=transformation_type, value=value, unit=unit, extra_metadata=extra_metadata
                )
            else:
                # DataField
                new_data_handler = AssonantDataHandlerFactory.create_data_field(
                    value=value, unit=unit, extra_metadata=extra_metadata
                )
        else:
            # It is a TimeSeries monitored ...
            if transformation_type is not None:
                # Axis
                new_data_handler = AssonantDataHandlerFactory.create_timeseries_axis(
                    transformation_type=transformation_type,
                    value=value,
                    timestamps=timestamps,
                    unit=unit,
                    extra_metadata=extra_metadata,
                    timestamps_unit=timestamps_unit,
                    timestamp_extra_metadata=timestamp_extra_metadata,
                )
            else:
                # DataField
                new_data_handler = AssonantDataHandlerFactory.create_timeseries_field(
                    value=value,
                    timestamps=timestamps,
                    unit=unit,
                    extra_metadata=extra_metadata,
                    timestamps_unit=timestamps_unit,
                    timestamp_extra_metadata=timestamp_extra_metadata,
                )

        return new_data_handler

    @staticmethod
    def create_data_field(
        value: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ) -> DataField:
        """Create new DataField handler.

        Args:
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to field
            collected data.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]], optional): Dict
            containing any aditional metadata related to collected data. Defaults to {}.

        Raises:
            AssonantDataHandlerFactoryException: Failure during objects creation.

        Returns:
            DataField: New DataField object handlind given values.
        """
        try:
            new_field = DataField(
                value=value,
                unit=unit,
                extra_metadata=extra_metadata,
            )
        except Exception as e:
            raise AssonantDataHandlerFactoryError("Failed to create new DataField handler.") from e

        return new_field

    @staticmethod
    def create_axis(
        transformation_type: TransformationType,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ) -> Axis:
        """Create new Axis handler.

        Args:
            transformation_type (TransformationType): Type of transformation done by axis.
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to axis
            collected data.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing any aditional metadata related to collected data. Defaults to {}.

        Raises:
            AssonantDataHandlerFactory: Failure during objects creation.

        Returns:
            Axis: New Axis object handling given data.
        """
        try:
            new_axis = Axis(
                transformation_type=transformation_type,
                value=AssonantDataHandlerFactory.create_data_field(
                    value=value,
                    unit=unit,
                    extra_metadata=extra_metadata,
                ),
            )
        except Exception as e:
            raise AssonantDataHandlerFactoryError("Failed to create new Axis handler") from e

        return new_axis

    @staticmethod
    def create_timeseries_field(
        value: Union[int, float, str, List, Type[np.ndarray], None],
        timestamps: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
        timestamps_unit: Optional[str] = None,
        timestamp_extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ) -> TimeSeries:
        """Create new Timeseries handler to handle timestamped DataField data.

        Args:
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to field
            collected data.
            timestamps (Union[int, float, str, List, Type[np.ndarray], None]): Timestamps related to data collected
            from the field.
            unit (Optional[str], optional): Measurement unit related to value parameter.
            Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing any aditional metadata related to collected data. Defaults to {}.
            timestamps_unit (Optional[str], optional): Measurement unit related to
            timestamp field. Defaults to None.
            timestamp_extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing extra metadata about timestamps field. Defaults to {}.

        Raises:
            AssonantDataHandlerFactoryException: Failure during objects creation.

        Returns:
            TimeSeries: New TimeSeries object handling given timestamped data.
        """
        try:
            new_field = TimeSeries(
                value=AssonantDataHandlerFactory.create_data_field(
                    value=value,
                    unit=unit,
                    extra_metadata=extra_metadata,
                ),
                timestamps=AssonantDataHandlerFactory.create_data_field(
                    value=timestamps,
                    unit=timestamps_unit,
                    extra_metadata=timestamp_extra_metadata,
                ),
            )
        except Exception as e:
            raise AssonantDataHandlerFactoryError("Failed to create new TimeSeries handler.") from e

        return new_field

    @staticmethod
    def create_timeseries_axis(
        transformation_type: TransformationType,
        value: Union[int, float, str, List, Type[np.ndarray], None],
        timestamps: Union[int, float, str, List, Type[np.ndarray], None],
        unit: Optional[str] = None,
        extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
        timestamps_unit: Optional[str] = None,
        timestamp_extra_metadata: Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray], None]]] = {},
    ) -> Axis:
        """Create new TimeSeries handler to handle timestamped Axis data.

        Args:
            transformation_type (TransformationType): Transformation type of the Axis
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value related to
            axis collected data
            timestamps (Union[int, float, str, List, Type[np.ndarray], None]): Timestamps
            related to data collected from the axis.
            unit (Optional[str], optional): Measurement unit related to value
            field. Defaults to None.
            extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dictionary
            containing extra metadata about value field. Defaults to {}.
            tracked it as a TimeSeries. Defaults to None.
            timestamps_unit (Optional[str], optional): Measurement unit related to
            timestamp field. Defaults to None.
            timestamp_extra_metadata (Optional[Dict[str, Union[int, float, str, List, Type[np.ndarray],
            None]]], optional): Dict
            containing extra metadata about timestamps field. Defaults to {}.

        Raises:
            AssonantDataHandlerFactoryException: ailure during objects creation.

        Returns:
            Axis: New Axis object handling given timestamped data.
        """
        try:
            new_axis = Axis(
                transformation_type=transformation_type,
                value=AssonantDataHandlerFactory.create_timeseries_field(
                    value=value,
                    unit=unit,
                    extra_metadata=extra_metadata,
                    timestamps=timestamps,
                    timestamps_unit=timestamps_unit,
                    timestamp_extra_metadata=timestamp_extra_metadata,
                ),
            )
        except Exception as e:
            raise AssonantDataHandlerFactoryError("Failed to create Axis DataHandler with TimeSeries data.") from e

        return new_axis



"""Tests focused on validating AssonantDataRetriever methods"""

import json
import os

import pytest

from assonant.data_retriever import AssonantDataRetriever
from assonant.data_retriever.exceptions import AssonantDataRetrieverError
from assonant.naming_standards import AcquisitionType, BeamlineName


def assert_retrieved_pv_data_structure(retrieved_pv_data):
    """
    Helper function to validate the structure of the retrieved PV data.

    Expected Structure:
        - The retrieved PV data must be a dictionary where:
            - Keys are strings representing PV names.
            - Values must always be returned inside lists as treatment must be done on highler layers.
    """
    assert isinstance(retrieved_pv_data, dict)

    for key, value in retrieved_pv_data.items():
        assert isinstance(key, str)
        # Data retriever must always return info within a list, treatment must be done on higher layers.
        assert isinstance(value, list)


@pytest.fixture(scope="session")
def example_bluesky_data_retriever(example_bluesky_all_streams_acquisition_file_path) -> AssonantDataRetriever:
    """
    Provides a functional AssonantMetadataRetriever for the example_metadata.csv.
    """
    return AssonantDataRetriever(data_source=example_bluesky_all_streams_acquisition_file_path)


def test_data_retriever_invalid_file_format():
    """
    Validate that AssonantDataRetriever constructor correctly raises exception
    if path to a file of an unsupported file format is passed during object construction.

    Expected Behavior:
        - AssonantDataRetriever must raise an DataRetrieverFactoryError if an filepath for unsupported
        file format is passed.
    """
    with pytest.raises(AssonantDataRetrieverError):
        _ = AssonantDataRetriever(data_source="UNSUPPORTED_FILE_FORMAT")


@pytest.mark.parametrize("acquisition_type", [acq_type for acq_type in AcquisitionType])
def test_get_pv_data_by_acquisition_type_returned_structure(example_bluesky_data_retriever, acquisition_type):
    """
    Validate that AssonantDataRetriever get_pv_data_by_acquisition_type() method correctly return expected
    value when acquisition type does exist.

    Expected Behavior:
        - AssonantDataRetriever be able to return a dictionary containing retrieved PV data
        from the acquisition type.
        - The dictionary structure must follow exactly the one define at the
        IAssonantDataRetriever interface.
    """
    retrieved_pv_data = example_bluesky_data_retriever.get_pv_data_by_acquisition_type(acquisition_type)

    assert_retrieved_pv_data_structure(retrieved_pv_data)


@pytest.mark.parametrize("beamline_name", [beamline_name for beamline_name in BeamlineName])
@pytest.mark.parametrize("acquisition_type", [acq_type for acq_type in AcquisitionType])
def test_get_pv_data_by_acquisition_type_returned_structure_over_real_example_data(
    beamline_name, acquisition_type, real_bluesky_data_test_cases_dir_path_map
):
    """
    Validate that AssonantDataRetriever get_pv_data_by_acquisition_type() method correctly return expected
    structure over real example data.

    Expected Behavior:
        - AssonantDataRetriever be able to return a dictionary containing retrieved PV data
        for each acquisition Type for all real data test cases that exist for each beamline.
        - The dictionary structure must follow exactly the one define at the
        IAssonantDataRetriever interface.
    """
    if beamline_name.value not in real_bluesky_data_test_cases_dir_path_map.keys():
        pytest.skip(reason=f"There is no test case with real data for {beamline_name} beamline yet!")
    else:

        for i in range(0, len(real_bluesky_data_test_cases_dir_path_map[beamline_name.value])):
            # Execute test for each test case dir that exist for that beamline
            test_case_dir = real_bluesky_data_test_cases_dir_path_map[beamline_name.value][i]

            for test_case_file in os.scandir(test_case_dir):
                if test_case_file.is_file() and test_case_file.name.endswith(".json"):

                    with open(test_case_file) as f:
                        test_case_bluesky_docs_list = json.load(f)

                data_retriever = AssonantDataRetriever(data_source=test_case_bluesky_docs_list)

                retrieved_pv_data = data_retriever.get_pv_data_by_acquisition_type(acquisition_type)
                assert_retrieved_pv_data_structure(retrieved_pv_data)



"""Assonant data classes - Data handlers submodule.

Data classes responsible for standardizing how data is handled inside other
Assonant data classes.
"""

from .axis import Axis
from .data_field import DataField
from .data_handler import DataHandler
from .time_series import TimeSeries

__all__ = [
    "Axis",
    "DataField",
    "DataHandler",
    "TimeSeries",
]



from enum import Enum


class CSVColumn(Enum):
    """CSV column values."""

    NAME = "Name"
    SUBCOMPONENT_OF = "Subcomponent of"
    CLASS = "Class"
    FIELD_NAME = "Field Name"
    UNIT_OF_MEASUREMENT = "Unit of Measurement"
    DATA_TYPE = "Data Type"
    ACQUISITION_TYPE = "Acquisition Type"
    PV_NAME = "PV Name"
    NEXUS_CLASS = "NeXus Class"
    NEXUS_FIELD_NAME = "NeXus Field Name"
    TRANSFORMATION_TYPE = "Transformation Type"



"""Assonant data classes enums.

This submodule defines Enumations classes used to standardize options from some data classes.
"""

from .transformation_type import TransformationType
from .value_placeholders import ValuePlaceholders

__all__ = ["TransformationType", "ValuePlaceholders"]



"""Exceptions triggered to warn about internal errors in Metadata Retriever Factory module."""


class MetadataRetrieverFactoryError(Exception):
    """Metadata Retriever Factory Error.

    Custom Exception to warn when any error occurs inside MetadataRetrieverFactory
    """

    pass



"""Tests focused on validating AssonantComponentFactory methods from Data Classes submodule"""

import importlib
import inspect

import pytest

from assonant.data_classes import AssonantComponentFactory
from assonant.data_classes.components import *
from assonant.data_classes.exceptions import AssonantComponentFactoryError


def test_create_component_by_class_name():
    """
    Validate that AssonantComponentFactory correctly instantiates components using their class names.

    Expected Behavior:
        - For every valid class defined in `assonant.data_classes.components`, the factory should:
            - Dynamically resolve the class name.
            - Instantiate the component with the provided name.
            - Return an instance of the expected component class with the correct 'name' attribute.
    """
    for name, cls in inspect.getmembers(importlib.import_module("assonant.data_classes.components"), inspect.isclass):
        component = AssonantComponentFactory.create_component_by_class_name(name, "component_name")
        assert isinstance(component, cls)
        assert component.name == "component_name"


def test_create_component_by_class_name_invalid():
    """
    Validate that AssonantComponentFactory raises an exception for unknown class names.

    Expected Behavior:
        - If the given class name does not exist in the component module,
          the factory should raise an AssonantComponentFactoryError to prevent silent failures.
    """
    with pytest.raises(AssonantComponentFactoryError):
        AssonantComponentFactory.create_component_by_class_name("NonExistentComponent", "bad")


def test_create_component_by_class_type():
    """
    Validate that AssonantComponentFactory correctly instantiates components using their class types.

    Expected Behavior:
        - For every valid component class, the factory should:
            - Accept the class type as input.
            - Instantiate the component using a compatible constructor.
            - Assign the given name to the 'name' attribute.
            - Return a properly typed instance of the component.
    """
    for name, cls in inspect.getmembers(importlib.import_module("assonant.data_classes.components"), inspect.isclass):
        component = AssonantComponentFactory.create_component_by_class_type(cls, "component_name")
        assert isinstance(component, cls)
        assert component.name == "component_name"


def test_create_component_by_class_type_invalid():
    """
    Validate that AssonantComponentFactory raises an error when the class constructor is incompatible.

    Expected Behavior:
        - If the class does not support instantiation with a single `name` argument,
          the factory should detect this and raise an AssonantComponentFactoryError
          to signal improper usage or a misconfigured component class.
    """

    class FakeComponent(Component):
        def __init__(self, fail):
            pass  # wrong signature

    with pytest.raises(AssonantComponentFactoryError):
        AssonantComponentFactory.create_component_by_class_type(FakeComponent, "broken")


def test_create_component_methods_consistency():
    """
    Validate that both AssonantComponentFactory creation methods
    return equal Component instances. That guarantee consistency among creational
    methods.

    Expected Behavior:
        - Both factory methods must return equivalent Component objects.
    """
    for name, cls in inspect.getmembers(importlib.import_module("assonant.data_classes.components"), inspect.isclass):
        component_created_with_class_name = AssonantComponentFactory.create_component_by_class_name(
            name, "component_name"
        )
        component_created_with_class_type = AssonantComponentFactory.create_component_by_class_type(
            cls, "component_name"
        )

        assert component_created_with_class_type == component_created_with_class_name



import pytest
from nexusformat.nexus import NXroot

from assonant._nexus import NeXusObjectFactory, NeXusObjectFactoryError
from assonant.data_classes import AssonantDataClass
from assonant.data_classes.components import Component
from assonant.data_classes.data_handlers import Axis, DataField, TimeSeries

from ..utils import get_component_classes, get_data_handler_classes


@pytest.mark.parametrize("component_cls", get_component_classes())
def test_create_nxobject_matches_component_mapping(component_cls, component_to_nexus_mapping):
    """Test that NeXusObjectFactory creates the correct NXobject for each component."""
    factory = NeXusObjectFactory()
    data_obj: AssonantDataClass = component_cls(name="test_component")

    nxobj = factory.create_nxobject(data_obj)

    # Check expected NeXus class was used for object
    expected_nx_class = component_to_nexus_mapping[component_cls]
    assert isinstance(nxobj, expected_nx_class)


@pytest.mark.parametrize("data_handler_cls", get_data_handler_classes())
def test_create_nxobject_matches_data_handler_mapping(
    data_handler_cls,
    data_handler_to_nexus_mapping,
    example_string_data_field,
    example_translation_axis,
    example_int_1d_timeseries_data_field,
    example_timeseries_rotation_axis,
):
    """Test that NeXusObjectFactory creates the correct NXobject for each component."""
    factory = NeXusObjectFactory()

    if data_handler_cls is DataField:
        nxobj = factory.create_nxobject(example_string_data_field)

        expected_nx_class = data_handler_to_nexus_mapping[data_handler_cls]
        assert isinstance(nxobj, expected_nx_class)
    elif data_handler_cls is Axis:
        # Axis is a specific case where it may have an DataField within it or an TimeSeries, depending on it returned
        # nexus object is different
        nxobj = factory.create_nxobject(example_translation_axis)
        nxobj2 = factory.create_nxobject(example_timeseries_rotation_axis)

        expected_nx_classes = data_handler_to_nexus_mapping[data_handler_cls]
        assert type(nxobj) in expected_nx_classes
        assert type(nxobj2) in expected_nx_classes

    elif data_handler_cls is TimeSeries:

        nxobj = factory.create_nxobject(example_int_1d_timeseries_data_field)

        expected_nx_class = data_handler_to_nexus_mapping[data_handler_cls]
        assert isinstance(nxobj, expected_nx_class)
    else:
        raise AssertionError(f"Test for class {data_handler_cls} was not developed!")


def test_create_nxobject_pack_into_nxroot(component_to_nexus_mapping, example_components_collection):
    """Test that objects can be packed into an NXroot when requested."""
    factory = NeXusObjectFactory()
    data_obj = example_components_collection["Detector"]

    nxobj = factory.create_nxobject(data_obj, pack_into_nxroot=True)

    # Check first NeXus class is the NXroot
    assert isinstance(nxobj, NXroot)

    # Check that component NeXus Object was created within NXroot
    assert data_obj.name in nxobj
    assert isinstance(nxobj[data_obj.name], component_to_nexus_mapping[type(data_obj)])


def test_create_nxobject_invalid_component():
    """Test that creating an NXobject from an unsupported component raises an error."""

    class FakeComponent(Component):
        """A dummy component not present in the mapping."""

        pass

    factory = NeXusObjectFactory()
    fake_obj = FakeComponent(name="fake_component")

    with pytest.raises(NeXusObjectFactoryError):
        factory.create_nxobject(fake_obj)



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union

from assonant.naming_standards import AcquisitionType


class IAssonantMetadataRetriever(ABC):
    """Assonant Metadata Retriever Interface.

    This interface defines what must be implemented by a class to be a Metadata Retriever
    in Assonant environment.
    """

    @abstractmethod
    def get_pv_names_by_acquisition_type(self, acquisition_type: AcquisitionType) -> List[str]:
        """Return a List with all PV names related to the passed AcquisitionType.

        Returned Structure Definition:
        [
            PV_1_NAME,
            PV_2_NAME,
            PV_3_NAME,
            ...
            PV_N_NAME
        ]

        Args:
            acquisition_type (AcquisitionType): Target Acquisition Type used for selecting which
            PV names will be returned.

        Raises:
            AssonantMetadataRetrieverError: Failure to retrieve requested metadata.

        Returns:
            List[str]: List containing all PV names related to passed AcquisitionType.
        """

    @abstractmethod
    def get_pvs_info(self, pv_names: Union[str, List[str]]) -> Dict[str, Dict[str, Any]]:
        """Return all information about the PV paired by the its name.

        Example of return structure.:
        {
            PV_1_NAME_: {
                component_info:
                {
                    name: COMPONENT_NAME,
                    class: COMPONENT_CLASS,
                    subcomponent_of: ASCENDANT_COMPONENT_NAME (if exists)
                },
                data_handler_info:
                {
                    name: FIELD_NAME,
                    value: PLACEHOLDER_VALUE,
                    pv_name: PV_1_NAME,
                    unit: FIELD_UNIT (if exists),
                    transformation_type: FIELD_TRANSFORMATION_TYPE (if exists),
                    PV_1_EXTRA_METADATA_1_NAME: PV_1_EXTRA_METADATA_1_VALUE (if exists),
                    PV_1_EXTRA_METADATA_2_NAME: PV_1_EXTRA_METADATA_2_VALUE (if exists),
                    ...
                }
            },
            PV_2_NAME_: {
                component_info:
                {
                    name: COMPONENT_NAME,
                    class: COMPONENT_CLASS,
                    subcomponent_of: ASCENDANT_COMPONENT_NAME (if exists),
                },
                data_handler_info:
                {
                    name: FIELD_NAME,
                    value: PLACEHOLDER_VALUE,
                    pv_name: PV_2_NAME,
                    unit: FIELD_UNIT (if exists),
                    transformation_type: FIELD_TRANSFORMATION_TYPE (if exists),
                    PV_2_EXTRA_METADATA_1_NAME: PV_2_EXTRA_METADATA_1_VALUE (if exists),
                    PV_2_EXTRA_METADATA_2_NAME: PV_2_EXTRA_METADATA_2_VALUE (if exists),
                    ...
                }
            },
        }

        Args:
            pv_names (List[str]): List with PV names which metadata info will be fetched.

        Raises:
            AssonantMetadataRetrieverError: Failure to retrieve requested metadata.

        Returns:
            Dict[str, Dict[str, Any]]: Dictionary containing all PV info paired by the PV name.
        """

    @abstractmethod
    def get_subcomponents_mapping(self) -> Dict[str, List[str]]:
        """Return a Dictionary mapping component names with the name of its subcomponents.

        This call MUST be indepedent of the AcquisitionType as components hierarchy is not related to
        acquisition directly but logical organization. If AcquisitionType is considered here, it may
        generate hierarchy inconsistency due to the possibility of parent Component exist in some
        AcquisitionTypes and not in others.

        Important considerations about the returned structure:
            * The returned structure contains only Components that have at least one Subcomponent.
            * Subcomponent can contain Subcomponents, that said, names that appears as Subcomponents
            for a Component may also appear as Component with other Subcomponents. Treating how to
            deal with this is responsibility of who is calling this method.

        Example of returned structure:
        {
            COMPONENT_1_NAME: [SUBCOMPONENT_1_NAME, SUBCOMPONENT_2_NAME, ...],
            COMPONENT_2_NAME: [SUBCOMPONENT_1_NAME, SUBCOMPONENT_2_NAME, ...],
            ...
        }

        Raises:
            AssonantMetadataRetrieverError: Failure to retrieve requested metadata.

        Returns:
            Dict[str, List[str]]: Dictionary mapping the name of each Component that contains at
            least one Subcomponent with their respective List of Subcomponents names.
        """

    @abstractmethod
    def get_component_info(self, component_name: str) -> Dict[str, Dict[str, Any]]:
        """Retrieve info about a specific Component based on passed component_name.

        Example of returned structure:
            CHECK get_pvs_info() 'component_info' structure. IT MUST BE THE SAME.

        Args:
            component_name (str): Component name to retrieve information about.

        Raises:
            AssonantMetadataRetrieverError: Failure to retrieve requested data.

        Returns:
            Dict[str, Dict[str, Any]]: Component info dictionary following the same subdictionary
            returned by get_pvs_info() method to represent component info.
        """



"""Assonant TimeSeries class."""

from typing import List, Type, Union

import numpy as np

from .data_field import DataField
from .data_handler import DataHandler


class TimeSeries(DataHandler):
    """Data class to handle any type of time series data."""

    value: DataField
    timestamps: DataField

    def set_value(self, value: Union[int, float, str, List, Type[np.ndarray], None]):
        """Set value to 'value' property from TimeSeries.

        Args:
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value to bet set to
            TimeSeries 'value' property.
        """
        self.value.set_value(value)

    def set_timestamps(self, value: Union[int, float, str, List, Type[np.ndarray], None]):
        """Set value to 'timestamps' property from TimeSeries.

        Args:
            value (Union[int, float, str, List, Type[np.ndarray], None]): Value to bet set to
            TimeSeries 'timestamps' property.
        """
        self.timestamps.set_value(value)



"""Assonant File Writer.

Assonant File Writer module handle all process related to writing AssonantDataClasses data into
files in all supported file formats implemented.
"""

from .assonant_file_writer import AssonantFileWriter
from .assonant_file_writer_interface import IAssonantFileWriter
from .exceptions import AssonantFileWriterError
from .file_writer_factory import FileWriterFactory

__all__ = ["AssonantFileWriter", "AssonantFileWriterError", "FileWriterFactory", "IAssonantFileWriter"]



import json
from typing import Any, Dict, List, Union

from event_model import DocumentRouter

from .exceptions import BlueskyDataParserError


class DocumentsDataParser(DocumentRouter):
    def __init__(self, bluesky_documents_list: List[List[Union[str, Dict[str, Any]]]], *args, **kwargs):
        """Constructor to instantiate a Bluesky Documents DataParser.

        Args:
            bluesky_documents_list (List[List[Union[str, Dict[str, Any]]]]): List of bluesky documents data
            presented also as a list. The list with the document data has 2 positions, the 1st is the bluesky
            identification for the document (start, event, stop, descriptor, ...) and the 2nd position the
            dict containing the document data.
        Raises:
            BlueskyDataParserError: Failure to read bluesky documents.
        """
        super().__init__(*args, **kwargs)

        self._start_document = None
        self._descriptor_documents = {}

        # Stream name --> Subdevice configurations
        self._configurations = {}
        # Stream name --> Device data keys
        self._data_keys = {}
        # Stream name --> Device data key --> (Device data reading, Reading timestamp)
        self._event_data = {}

        try:
            # print("Init DocumentsDataParser")
            # Iterate over json data correctly parse data based on event type
            for event_name, document_data_dict in bluesky_documents_list:
                # print(event_name, document_data_dict)
                self(event_name, document_data_dict)
        except Exception as e:
            raise BlueskyDataParserError(
                f"Failed to read bluesky documents from structure: {bluesky_documents_list}"
            ) from e

    def start(self, doc: Dict):
        """DocumentRouter method related to processing bluesky START documments

        Args:
            doc (Dict): Dictionary containing bluesky documment data.
        """
        # print("========== START ==========")
        # print(doc)
        # print("===========================\n")
        self._start_document = doc

    def descriptor(self, doc: Dict):
        """DocumentRouter method related to processing bluesky DESCRIPTOR documments

        Args:
            doc (Dict): Dictionary containing bluesky documment data.
        """
        # print("========== DESC ==========")
        # print(doc)
        # print("===========================\n")
        self._descriptor_documents[doc["uid"]] = doc

        self._configurations[doc["name"]] = doc["configuration"]
        self._data_keys[doc["name"]] = doc["data_keys"]
        # print("DESC DOCS")
        # print(self._descriptor_documents)

    def event(self, doc: Dict):
        """DocumentRouter method related to processing bluesky EVENT documments

        Args:
            doc (Dict): Dictionary containing bluesky documment data.
        """
        # print("========== EVENT ==========")
        # print(doc)
        # print("===========================\n")
        stream_name = self._descriptor_documents[doc["descriptor"]]["name"]

        if stream_name not in self._event_data:
            self._event_data[stream_name] = {}

        if doc["seq_num"] not in self._event_data[stream_name]:
            self._event_data[stream_name][doc["seq_num"]] = {}

        for dev_name, val in doc["data"].items():
            self._ensure_dict_key(dev_name, self._event_data[stream_name][doc["seq_num"]])

            self._event_data[stream_name][doc["seq_num"]][dev_name] = {
                "value": val,
                "timestamps": doc["timestamps"][dev_name],
            }

    def get_stream_names(self) -> List[str]:
        """Return a List of streams names within parsed documents.

        Returns:
            List[str]: List of stream names found on parsed documents.
        """
        return [stream_name for stream_name in self._data_keys]

    def all_data_keys(self) -> Dict[str, List]:
        """Return all dictionary keys to access data fields from each stream into the data dictionary.

        PS: If no device_data_key is present for a stream, it will return an empty list for it.

        Returns:
            Dict[str, List]: Dictionary containing the stream name and a List with its respetive data_keys
        """
        data_keys = {}
        for stream, devices in self._data_keys.items():
            # Even if stream has no device within it, return an empty list for it
            if stream not in data_keys:
                data_keys[stream] = []
            for device_data_key, device_values in devices.items():
                data_keys[stream].append(device_data_key)
        return data_keys

    def data_key_to_pv_name(self, data_key: str) -> str:
        """Convert data_key to the respective PV name related to it.

        Args:
            data_key (str): data key from data dictionary that will be converted

        Returns:
            str: PV Name respective to passed data_key
        """
        pv_names = {}
        for stream, devices in self._data_keys.items():
            for device_data_key, device_values in devices.items():
                if device_data_key == data_key:
                    pv_names[stream] = ":".join(device_values["source"].split(":")[1:])
                    break
        return pv_names

    def _ensure_dict_key(self, key_name: Any, target_dict: Dict) -> None:
        """
        Ensures that a specified key exists in the given dictionary.
        If the key does not exist, it initializes it with an empty dictionary.

        Args:
            key_name (Any): The key to ensure exists in the dictionary.
            target_dict (Dict): The dictionary to check and modify.
        """
        if key_name not in target_dict:
            target_dict[key_name] = {}

    def config_data_key_to_pv_name(self, data_key: str) -> str:
        """Convert a configuration data_key to the respective PV name related to it.

        Args:
            data_key (str): data key from configuration dictionary from descriptor event
            that will be converted

        Returns:
            str: PV Name respective to passed configuration data_key
        """
        pv_names = {}
        for stream, devices in self._configurations.items():
            for device_name in devices:
                for config_data_key, config_values in devices[device_name]["data_keys"].items():
                    if config_data_key == data_key:
                        pv_names[stream] = ":".join(config_values["source"].split(":")[1:])
                        break
        return pv_names

    def pv_name_to_data_key(self, pv_name: str) -> str:
        """Convert PV name to the respective data_key related to it.

        Args:
            pv_name (str): PV Name that will be converted

        Returns:
            str: Data key respective to passed PV Name
        """
        data_keys = {}
        formatted_pv_name = f"PV:{pv_name}"
        for stream, devices in self._data_keys.items():
            for device_data_key, device_values in devices.items():
                if formatted_pv_name in device_values["source"]:
                    data_keys[stream] = device_data_key
                    break

        return data_keys

    def pv_name_to_config_data_key(self, pv_name: str) -> str:
        """Convert PV name to the respective configuration data_key related to it.

        Args:
            pv_name (str): PV Name that will be converted

        Returns:
            str: Configuration Data key respective to passed PV Name
        """
        data_keys = {}
        formatted_pv_name = f"PV:{pv_name}"
        for stream, devices in self._configurations.items():
            for device_name in devices:
                for config_data_key, config_values in devices[device_name]["data_keys"].items():
                    if formatted_pv_name in config_values["source"]:
                        data_keys[stream] = config_data_key
                        break

        return data_keys

    def query_config_by_pv_name(self, pv_name: str, stream_name: str) -> Dict[str, Dict[str, Any]]:
        """Return configuration information related to a PV name.

        The returned dictionary structure is the following:
            {
                CONFIG_PV_NAME_1: {
                        "value": CONFIG_VALUE,
                        "timestamps": CONFIG_VALUE_ACQ_TIMESTAMP
                },
                CONFIG_PV_NAME_2: {
                        "value": CONFIG_VALUE,
                        "timestamps": CONFIG_VALUE_ACQ_TIMESTAMP
                    }
                    ...
            }

        Args:
            pv_name (str): PV name that will be used to query configuration information.
            stream_name (str): Stream name that will be used to query configuration information.

        Returns:
            Dict[str, Dict[str, Any]]: Dictionary containing configuration information related to passed PV name
            structured as described above. If no config is found it returns an empty dict.

        """

        config_data = {}

        if stream_name not in self._configurations:
            return config_data

        for device_name in self._configurations[stream_name]:
            for config_data_key, config_values in self._configurations[stream_name][device_name]["data_keys"].items():
                config_pv_name = ":".join(config_values["source"].split(":")[1:])

                self._ensure_dict_key(config_pv_name, config_data)

                config_data[config_pv_name]["value"] = self._configurations[stream_name][device_name]["data"][
                    config_data_key
                ]
                config_data[config_pv_name]["timestamps"] = self._configurations[stream_name][device_name][
                    "timestamps"
                ][config_data_key]

        return config_data

    def query_data_by_stream_name(
        self, event_stream_name: str, with_config: bool = True
    ) -> Dict[int, Dict[str, Dict[str, Any]]]:
        """Return all information from bluesky docs from a specific event stream as described below.

        The organization structure returned is the following:
        {
            SEQ_NUM_1: {
                PV_NAME_A: {
                    "data": {"value": PV_VALUE_1, "timestamps": PV_VALUE_1_TIMESTAMP},
                    "configs": <Structure from query_config_by_pv_name() method>
                },
                PV_NAME_B: {
                    "data": {"value": PV_VALUE_1, "timestamps": PV_VALUE_1_TIMESTAMP},
                    "configs": <Structure from query_config_by_pv_name() method>
                },
            },
            SEQ_NUM_2: {
                PV_NAME_A: {
                        "data": {"value": PV_VALUE_2, "timestamps": PV_VALUE_2_TIMESTAMP},
                        "configs": <Structure from query_config_by_pv_name() method>
                    },
                    PV_NAME_B: {
                        "data": {"value": PV_VALUE_2, "timestamps": PV_VALUE_2_TIMESTAMP},
                        "configs": <Structure from query_config_by_pv_name() method>
                    },
        }

        PS: If stream exist but no device data is found, it returns an empty dict.
        PS: PV values may be list

        Args:
            event_stream_name (str): Query's target event stream name.
            with_config (bool): Flag to control if data will add configuration
            information or not. Defaults to True.

        Returns:
            Dict[int, Dict[str, Dict[str, Union[List[Dict[str, Any]], Dict[str, Any]]]]]: Dictionary
            with data following structure described above. If no data is found it returns an empty dict.
        """
        # Stream name --> Seq_num --> PV prefix --> Collected data
        # data = {}

        data = {}

        data_keys = self.all_data_keys()
        if len(data_keys) == 0 or event_stream_name not in data_keys.keys():
            return data

        data_keys = data_keys[event_stream_name]

        for data_key in data_keys:
            pv_name = self.data_key_to_pv_name(data_key)[event_stream_name]

            device_data = self._event_data.get(event_stream_name, {})
            for seq_num in device_data:

                self._ensure_dict_key(seq_num, data)
                self._ensure_dict_key(pv_name, data[seq_num])
                data[seq_num][pv_name]["data"] = device_data[seq_num].get(data_key, {})

                if with_config:
                    data[seq_num][pv_name]["configs"] = self.query_config_by_pv_name(pv_name, event_stream_name)

        return data

    def query_all_data(
        self, with_config: bool = True
    ) -> Dict[str, Dict[int, Dict[str, Dict[str, Union[List[Dict[str, Any]], Dict[str, Any]]]]]]:
        """Return all information from bluesky docs as described below.

        The organization structure returned is the following:
        {
            STREAM_NAME_1: <Structure from query_data_by_stream_name() method>,
            STREAM_NAME_2: <Structure from query_data_by_stream_name() method>,
            ...
        }

        PS: PV values are ALWAYS inside a List, even if it is a single value. It is also ALWAYS
        returned in a Tuple where the first position is the acquired value and the second the
        acquisition timestamp.

        Args:
            with_config (bool): Flag to control if data will add configuration
            information or not. Defaults to True.

        Returns:
            Dict[str, Dict[int, Dict[str, Dict[str, Union[List[Dict[str, Any]], Dict[str, Any]]]]]]: Dictionary
            with data following structure described above. If no data is found it returns an empty dict.
        """
        # Stream name --> structured from query_data_by_stream_name()
        data = {}

        data_keys = self.all_data_keys()
        if len(data_keys) == 0:
            return data

        for stream in data_keys.keys():
            # print(f"Stream '{stream}':")
            data[stream] = {}
            stream_data = self.query_data_by_stream_name(stream, with_config=with_config)
            data[stream] = stream_data
        return data

    def load_json(self, path: str) -> Dict[str, Any]:
        """Load data from passed json file

        Args:
            path (str): File path to json file containing data collect from bluesky.

        Returns:
            Dict[str, Any]: Dictionary containing retrieved data from json file.
        """
        contents = None
        with open(path, "r") as _f:
            contents = json.load(_f)

        return contents



"""Assonant data parsers for data loggers.

This submodule defines Data Parsers objects to be used on Data Logger in order to interpretate
different types of data mapping/configuration files.
"""

from .assonant_data_retriever import AssonantDataRetriever
from .assonant_data_retriever_interface import IAssonantDataRetriever
from .data_retriever_factory import DataRetrieverFactory

__all__ = ["AssonantDataRetriever", "DataRetrieverFactory", "IAssonantDataRetriever"]



"""Tests focused on validating Assonant Components/Entry methods that add information to it"""

import pytest

from assonant.data_classes.data_handlers import Axis, DataField, TimeSeries
from assonant.data_classes.enums import TransformationType
from assonant.data_classes.factories import AssonantDataHandlerFactory

from ..utils import (
    assert_axis_content,
    assert_axis_equality,
    assert_component_equality,
    assert_data_field_content,
    assert_data_fields_equality,
    assert_timeseries_content,
    assert_timeseries_equality,
    get_fixture_values,
)
from .test_data_handler_factory import create_test_params_combinations


@pytest.mark.parametrize("method", ["specific", "generic"])
@pytest.mark.parametrize("value, unit, metadata", create_test_params_combinations(DataField))
def test_add_field(value, unit, metadata, method, example_base_component, request):
    """
    Validate that add_field and add_data_handler methods from Components correctly
    add DataField object to its internal fields dictionary for all accepted data types.

    Expected Behavior:
        - Base Component must add DataField within its fields dictionary
        - Added DataField must be identified with the same name it was given during add_field method
        - DataField object within Component fields dictionary must be equivalent to how it was before insertion
    """

    # To avoid creating lots of different fixtures, manually create DataField DataHandler using Factory
    # if this test passes, other tests related to adding DataField, don't need to deal with all data types.
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)

    data_field = AssonantDataHandlerFactory.create_data_field(
        value=input_value,
        unit=input_unit,
        extra_metadata=input_metadata,
    )
    data_field_name = "data_field"

    if method == "specific":
        target_method_ref = example_base_component.add_field
    if method == "generic":
        target_method_ref = example_base_component.add_data_handler

    target_method_ref(data_field_name, data_field.model_copy(deep=True))

    # Assert DataField was add to fields dict
    assert len(example_base_component.fields.items()) == 1

    # Assert correct name was given to DataField within fields dict and that it remained unchanged.
    assert_data_fields_equality(example_base_component.fields[data_field_name], data_field)


@pytest.mark.parametrize("method", ["specific", "generic"])
def test_add_field_with_name_conflict(
    example_base_component, example_int_1d_array_data_field, example_float_2d_array_data_field, method
):
    """
    Validate that add_field and add_data_handler method from Components correctly add
    DataField object to its internal fields dictionary and handles name conflicts.

    Expected Behavior:
        - Both added DataFields must be within Component fields internal dict
        - The first added DataField must not suffer any change from how it was before
        - Any subsequent DataField added that has the same name with another already existing from the fields
        dict from Component must have its name automatically updated to a new unique name by adding an index in
        front the original passed name (e.g: <field_name> (1))
    """
    field_name = "field_name"

    if method == "specific":
        target_method_ref = example_base_component.add_field
    elif method == "generic":
        target_method_ref = example_base_component.add_data_handler

    target_method_ref(field_name, example_int_1d_array_data_field.model_copy(deep=True))
    target_method_ref(field_name, example_float_2d_array_data_field.model_copy(deep=True))

    # Assert both DataField objects were add to positions dict
    assert len(example_base_component.fields.items()) == 2

    # 1st DataField name and content must be unchanged
    assert_data_fields_equality(example_base_component.fields[field_name], example_int_1d_array_data_field)

    # 2nd DataField name must be the same as the 1st with a index due to name conflict and its content must be unchanged
    assert_data_fields_equality(example_base_component.fields[field_name + " (1)"], example_float_2d_array_data_field)


@pytest.mark.parametrize("value, unit, metadata", create_test_params_combinations(DataField))
def test_add_field_methods_consistency(value, unit, metadata, example_base_component, request):
    """Validate that both add_field and add_data_handler methods from Components result in
    the same outcome when adding DataField to Component.

    Expected Behavior:
        - Both added DataField must be within its respective fields internal dict
        - Both added DataField must be indentified with the same name within its respective fields internal dict
        - Both added DataField must be equivalent
    """

    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)

    data_field = AssonantDataHandlerFactory.create_data_field(
        value=input_value,
        unit=input_unit,
        extra_metadata=input_metadata,
    )
    data_field_name = "data_field"

    component_1 = example_base_component.model_copy(deep=True)
    component_2 = example_base_component

    component_1.add_field(data_field_name, data_field)
    component_2.add_data_handler(data_field_name, data_field)

    assert_data_fields_equality(component_1.fields[data_field_name], component_2.fields[data_field_name])


@pytest.mark.parametrize("value, unit, metadata", create_test_params_combinations(DataField))
def test_create_and_add_field(value, unit, metadata, example_base_component, request):
    """
    Validate that create_and_add_field method from Components correctly
    create and add DataField object to its internal fields dictionary for all accepted data types.

    Expected Behavior:
        - Base Component must create a new DataField with passed data
        - Base Component must add the new created DataField within its fields dictionary
        - Added DataField must be identified with the same name it was given during create_and_add_field method
        - DataField object created by the Component must preserve passed data values.
    """
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)

    data_field_name = "data_field"
    example_base_component.create_and_add_field(data_field_name, input_value, input_unit, input_metadata)

    # Assert created DataField was add to fields dict
    assert len(example_base_component.fields.items()) == 1

    # Assert correct name was given to DataField within fields dict and that it remained unchanged
    assert_data_field_content(example_base_component.fields[data_field_name], input_value, input_unit, input_metadata)


def test_create_and_add_field_with_name_conflict(
    example_float_1d_array, example_int_1d_array, example_unit, example_metadata, example_base_component
):
    """
    Validate that create_and_add_field method from Components correctly
    create and add DataField object to its internal fields dictionary and handles name conflict.

    Expected Behavior:
        - Both DataFields must be created
        - Both created DataFields must be add to Component fields dictionary
        - The first create and added DataField must have its name and content preserved
        - The second create and added DataField must have its content preserved but name updated
        to a new unique name by adding an index in front of the original anem (e.g: <data_field_name> (1))
    """
    data_field_name = "data_field"
    example_base_component.create_and_add_field(data_field_name, example_float_1d_array, example_unit, example_metadata)
    example_base_component.create_and_add_field(data_field_name, example_int_1d_array, example_unit, example_metadata)

    # Assert both created DataFields were add to fields dict
    assert len(example_base_component.fields.items()) == 2

    # 1st DataField name and content must be unchanged
    assert_data_field_content(
        example_base_component.fields[data_field_name], example_float_1d_array, example_unit, example_metadata
    )

    # 2nd DataField name must be the same as the 1st with an index due to name conflict and
    # its content must be unchanged
    assert_data_field_content(
        example_base_component.fields[data_field_name + " (1)"], example_int_1d_array, example_unit, example_metadata
    )


@pytest.mark.parametrize("method", ["specific", "generic"])
@pytest.mark.parametrize("transformation_type", [t_type for t_type in TransformationType])
@pytest.mark.parametrize("value, unit, metadata", create_test_params_combinations(Axis))
def test_add_position(value, unit, metadata, transformation_type, method, example_base_component, request):
    """
    Validate that add_position and add_data_handler method from Components correctly
    add Axis object to its internal positions dictionary for all accepted data types.

    Expected Behavior:
        - Base Component must add Axis within its position dictionary
        - Added Axis must be identified with the same name it was given during add_position method
        - Axis object within Component positions dictionary must be equivalent to how it was before insertion
    """

    # To avoid creating lots of different fixtures, manually create Axis DataHandler using Factory
    # if this test passes, other tests related to adding Axis, don't need to deal with all data types.
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)

    axis = AssonantDataHandlerFactory.create_axis(
        transformation_type=transformation_type,
        value=input_value,
        unit=input_unit,
        extra_metadata=input_metadata,
    )
    axis_name = "x"

    if method == "specific":
        target_method_ref = example_base_component.add_position
    if method == "generic":
        target_method_ref = example_base_component.add_data_handler

    target_method_ref(axis_name, axis.model_copy(deep=True))

    # Assert Axis was add to positions dict
    assert len(example_base_component.positions.items()) == 1

    # Assert correct name was given to Axis within positions dict and that it remained unchanged.
    assert_axis_equality(example_base_component.positions[axis_name], axis)


@pytest.mark.parametrize("method", ["specific", "generic"])
def test_add_position_with_name_conflict(
    example_base_component, example_translation_axis, example_rotation_axis, method
):
    """
    Validate that add_position and add_data_handler method from Components correctly add Axis object
    to its internal positions dictionary and handles name conflicts.

    Expected Behavior:
        - Both added Axis must be within Component positions internal dict
        - The first added Axis must not suffer any change from how it was before
        - Any subsequent Axis added that has the same name with another already existing Axis from the poisitions
        dict from Component must have its name automatically updated to a new unique name by adding an index in
        front the original passed name (e.g: <axis_name> (1))
    """
    axis_name = "x"

    if method == "specific":
        target_method_ref = example_base_component.add_position
    if method == "generic":
        target_method_ref = example_base_component.add_data_handler

    target_method_ref(axis_name, example_translation_axis.model_copy(deep=True))
    target_method_ref(axis_name, example_rotation_axis.model_copy(deep=True))

    # Assert both Axis objects were add to positions dict
    assert len(example_base_component.positions.items()) == 2

    # 1st Axis name and content must be unchanged
    assert_axis_equality(example_base_component.positions[axis_name], example_translation_axis)

    # 2nd Axis name must be the same as the 1st with an index due to name conflict and its content must be unchanged
    assert_axis_equality(example_base_component.positions[axis_name + " (1)"], example_rotation_axis)


@pytest.mark.parametrize("transformation_type", [t_type for t_type in TransformationType])
@pytest.mark.parametrize("value, unit, metadata", create_test_params_combinations(Axis))
def test_add_position_methods_consistency(value, unit, metadata, transformation_type, example_base_component, request):
    """Validate that both add_positions and add_data_handler methods from Components result in
    the same outcome when adding Axis to Component.

    Expected Behavior:
        - Both added Axis must be within its respective positions internal dict
        - Both added Axis must be indentified with the same name within its respective position internal dict
        - Both added Axis must be equivalent
    """

    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)

    axis = AssonantDataHandlerFactory.create_axis(
        transformation_type=transformation_type,
        value=input_value,
        unit=input_unit,
        extra_metadata=input_metadata,
    )
    axis_name = "x"

    component_1 = example_base_component.model_copy(deep=True)
    component_2 = example_base_component

    component_1.add_position(axis_name, axis)
    component_2.add_data_handler(axis_name, axis)

    assert_axis_equality(component_1.positions[axis_name], component_2.positions[axis_name])


@pytest.mark.parametrize("transformation_type", [t_type for t_type in TransformationType])
@pytest.mark.parametrize("value, unit, metadata", create_test_params_combinations(DataField))
def test_create_and_add_position(value, unit, metadata, transformation_type, example_base_component, request):
    """
    Validate that create_and_add_position method from Components correctly
    create and add Axis object to its internal positions dictionary for all accepted data types.

    Expected Behavior:
        - Base Component must create a new Axis with passed data
        - Base Component must add the new created Axis within its positions dictionary
        - Added Axis must be identified with the same name it was given during create_and_add_position method
        - Axis object created by the Component must preserve passed data values.
    """
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)
    input_transformation_type = transformation_type

    axis_name = "x"

    example_base_component.create_and_add_position(
        axis_name, input_transformation_type, input_value, input_unit, input_metadata
    )

    # Assert created Axis was add to positions dict
    assert len(example_base_component.positions.items()) == 1

    # Assert correct name was given to Axis within positions dict and that it remained unchanged
    assert_axis_content(
        example_base_component.positions[axis_name], input_value, input_unit, input_metadata, input_transformation_type
    )


def test_create_and_add_position_with_name_conflict(
    example_float_1d_array, example_int_1d_array, example_unit, example_metadata, example_base_component
):
    """
    Validate that create_and_add_position method from Components correctly
    create and add Axis object to its internal positions dictionary and handles name conflict.

    Expected Behavior:
        - Both Axis DataHandlers must be created
        - Both created Axis must be add to Component positions dictionary
        - The first create and added Axis must have its name and content preserved
        - The second create and added Axis must have its content preserved but name updated
        to a new unique name by adding an index in front of the original anem (e.g: <axis_name> (1))
    """
    axis_name = "x"
    example_base_component.create_and_add_position(
        axis_name, TransformationType.TRANSLATION, example_float_1d_array, example_unit, example_metadata
    )
    example_base_component.create_and_add_position(
        axis_name, TransformationType.ROTATION, example_int_1d_array, example_unit, example_metadata
    )

    # Assert both created Axis were add to positions dict
    assert len(example_base_component.positions.items()) == 2

    # 1st Axis name and content must be unchanged
    assert_axis_content(
        example_base_component.positions[axis_name],
        example_float_1d_array,
        example_unit,
        example_metadata,
        TransformationType.TRANSLATION,
    )

    # 2nd Axis name must be the same as the 1st with an index due to name conflict and its content must be unchanged
    assert_axis_content(
        example_base_component.positions[axis_name + " (1)"],
        example_int_1d_array,
        example_unit,
        example_metadata,
        TransformationType.ROTATION,
    )


@pytest.mark.parametrize("method", ["specific", "generic"])
@pytest.mark.parametrize("value, unit, metadata, timestamp", create_test_params_combinations(TimeSeries))
def test_add_timeseries_field(value, unit, metadata, timestamp, method, example_base_component, request):
    """
    Validate that add_timeseries_field and add_data_handler method from Components correctly
    add TimeSeries with DataField object to its internal fields dictionary for all accepted data types.

    Expected Behavior:
        - Base Component must add passed TimeSeries within its fields dictionary
        - Added TimeSeries must be identified with the same name it was given during add_timeseries_field method
        - TimeSeries object within Component fields dictionary must be equivalent to how it was before insertion
    """

    # To avoid creating lots of different fixtures, manually create DataField DataHandler using Factory
    # if this test passes, other tests related to adding DataField, don't need to deal with all data types.
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)
    input_timestamp = get_fixture_values(timestamp, request)

    timeseries_data_field = AssonantDataHandlerFactory.create_timeseries_field(
        value=input_value,
        timestamps=input_timestamp,
        unit=input_unit,
        extra_metadata=input_metadata,
        timestamps_unit=input_unit,
        timestamp_extra_metadata=input_metadata,
    )
    timeseries_data_field_name = "timeseries_data_field"

    if method == "specific":
        target_method_ref = example_base_component.add_timeseries_field
    if method == "generic":
        target_method_ref = example_base_component.add_data_handler

    target_method_ref(timeseries_data_field_name, timeseries_data_field.model_copy(deep=True))

    # Assert TimeSeries with DataField was add to fields dict
    assert len(example_base_component.fields.items()) == 1

    # Assert correct name was given to DataField within fields dict and that it remained unchanged.
    assert_timeseries_equality(example_base_component.fields[timeseries_data_field_name], timeseries_data_field)


@pytest.mark.parametrize("method", ["specific", "generic"])
def test_add_timeseries_field_with_name_conflict(
    example_base_component, example_int_1d_timeseries_data_field, example_float_2d_timeseries_data_field, method
):
    """
    Validate that add_timeseries_field and add_data_handler method from Components correctly add TimeSeries
    object to its internal fields dictionary and handles name conflicts.

    Expected Behavior:
        - Both added TimeSeries must be within Component fields internal dict
        - The first added TimeSeries must not suffer any change from how it was before
        - Any subsequent TimeSeries added that has the same name with another already existing from the fields
        dict from Component must have its name automatically updated to a new unique name by adding an index in
        front the original passed name (e.g: <timeseries_field_name> (1))
    """
    timeseries_field_name = "timeseries_field_name"

    if method == "specific":
        target_method_ref = example_base_component.add_timeseries_field
    if method == "generic":
        target_method_ref = example_base_component.add_data_handler

    target_method_ref(timeseries_field_name, example_int_1d_timeseries_data_field.model_copy(deep=True))
    target_method_ref(timeseries_field_name, example_float_2d_timeseries_data_field.model_copy(deep=True))

    # Assert both DataField objects were add to positions dict
    assert len(example_base_component.fields.items()) == 2

    # 1st TimeSeries name and content must be unchanged
    assert_timeseries_equality(
        example_base_component.fields[timeseries_field_name], example_int_1d_timeseries_data_field
    )

    # 2nd TimeSeries name must be the same as the 1st with a index due to name conflict and
    # its content must be unchanged
    assert_timeseries_equality(
        example_base_component.fields[timeseries_field_name + " (1)"], example_float_2d_timeseries_data_field
    )


@pytest.mark.parametrize("value, unit, metadata, timestamp", create_test_params_combinations(TimeSeries))
def test_add_timeseries_field_methods_consistency(value, unit, metadata, timestamp, example_base_component, request):
    """Validate that both add_timeseries_field and add_data_handler methods from Components result in
    the same outcome when adding TimeSeries with DataField value to Component.

    Expected Behavior:
        - Both added TimeSeries must be within its respective fields internal dict
        - Both added TimeSeries must be indentified with the same name within its respective fields internal dict
        - Both added TimeSeries must be equivalent
    """
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)
    input_timestamp = get_fixture_values(timestamp, request)

    timeseries_field = AssonantDataHandlerFactory.create_timeseries_field(
        value=input_value,
        timestamps=input_timestamp,
        unit=input_unit,
        extra_metadata=input_metadata,
        timestamps_unit=input_unit,
        timestamp_extra_metadata=input_metadata,
    )
    timeseries_field_name = "timeseries_field_name"

    component_1 = example_base_component.model_copy(deep=True)
    component_2 = example_base_component

    component_1.add_timeseries_field(timeseries_field_name, timeseries_field)
    component_2.add_data_handler(timeseries_field_name, timeseries_field)

    assert_timeseries_equality(component_1.fields[timeseries_field_name], component_2.fields[timeseries_field_name])


@pytest.mark.parametrize("value, unit, metadata, timestamp", create_test_params_combinations(TimeSeries))
def test_create_and_add_timeseries_field(value, unit, metadata, timestamp, example_base_component, request):
    """
    Validate that create_and_add_timeseries_field method from Components correctly
    add TimeSeries with DataField object to its internal fields dictionary for all accepted data types.

    Expected Behavior:
        - Base Component must create a new TimeSeries within passed data
        - Base Component must add the new created TimeSeries within its fields dictionary
        - Added TimeSeries must be identified with the same name it was given during
        create_and_add_timeseries_field method
        - TimeSeries object created by the Component must preserve passed data values.
    """
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)
    input_timestamp = get_fixture_values(timestamp, request)

    timeseries_data_field_name = "timeseries_data_field"
    example_base_component.create_and_add_timeseries_field(
        name=timeseries_data_field_name,
        value=input_value,
        timestamps=input_timestamp,
        unit=input_unit,
        extra_metadata=input_metadata,
        timestamps_unit=input_unit,
        timestamp_extra_metadata=input_metadata,
    )

    # Assert TimeSeries with DataField was add to fields dict
    assert len(example_base_component.fields.items()) == 1

    # Assert correct name was given to DataField within fields dict and that it remained unchanged.
    assert_timeseries_content(
        example_base_component.fields[timeseries_data_field_name],
        input_value,
        input_unit,
        input_metadata,
        input_timestamp,
        input_unit,
        input_metadata,
    )


def test_create_and_add_timeseries_field_with_name_conflict(
    example_base_component, example_float_1d_array, example_int_1d_array, example_unit, example_metadata
):
    """
    Validate that create_and_add_timeseries_field method from Components correctly
    create and add TimeSeries object to its internal fields dictionary and handles name conflict.

    Expected Behavior:
        - Both TimeSeries must be created
        - Both created TimeSeries must be add to Component fields dictionary
        - The first created and added TimeSeries must have its name and content preserved
        - The second created and added TimeSeries must have its content preserved but name updated
        to a new unique name by adding an index in front of the original anem (e.g: <timeseries_data_field_name> (1))
    """
    timeseries_data_field_name = "timeseries_data_field"
    example_base_component.create_and_add_timeseries_field(
        timeseries_data_field_name,
        example_float_1d_array,
        example_float_1d_array,
        example_unit,
        example_metadata,
        example_unit,
        example_metadata,
    )
    example_base_component.create_and_add_timeseries_field(
        timeseries_data_field_name,
        example_int_1d_array,
        example_int_1d_array,
        example_unit,
        example_metadata,
        example_unit,
        example_metadata,
    )

    # Assert both created TimeSeries were add to fields dict
    assert len(example_base_component.fields.items()) == 2

    # 1st TimeSeries name and content must be unchanged
    assert_timeseries_content(
        example_base_component.fields[timeseries_data_field_name],
        example_float_1d_array,
        example_unit,
        example_metadata,
        example_float_1d_array,
        example_unit,
        example_metadata,
    )

    # 2nd TimeSeries name must be the same as the 1st with an index due to name conflict and
    # its content must be unchanged
    assert_timeseries_content(
        example_base_component.fields[timeseries_data_field_name + " (1)"],
        example_int_1d_array,
        example_unit,
        example_metadata,
        example_int_1d_array,
        example_unit,
        example_metadata,
    )


@pytest.mark.parametrize("method", ["specific", "generic"])
@pytest.mark.parametrize("transformation_type", [t_type for t_type in TransformationType])
@pytest.mark.parametrize("value, unit, metadata, timestamp", create_test_params_combinations(TimeSeries))
def test_add_timeseries_position(
    value, unit, metadata, timestamp, transformation_type, method, example_base_component, request
):
    """
    Validate that add_position and add_data_handler method from Components correctly add Axis
    object with TimeSeries to its internal positions dictionary for all accepted data types.

    Expected Behavior:
        - Base Component must add Axis within its position dictionary
        - Added Axis must be identified with the same name it was given during add_position method
        - Axis object within Component positions dictionary must be equivalent to how it was before insertion
    """

    # To avoid creating lots of different fixtures, manually create Axis DataHandler with TimeSeries using Factory
    # if this test passes, other tests related to adding Axis, don't need to deal with all data types.
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)
    input_timestamp = get_fixture_values(timestamp, request)
    input_transformation_type = transformation_type

    timeseries_axis = AssonantDataHandlerFactory.create_timeseries_axis(
        transformation_type=input_transformation_type,
        value=input_value,
        timestamps=input_timestamp,
        unit=input_unit,
        extra_metadata=input_metadata,
        timestamps_unit=input_unit,
        timestamp_extra_metadata=input_metadata,
    )
    timeseries_axis_name = "x"

    if method == "specific":
        target_method_ref = example_base_component.add_position
    if method == "generic":
        target_method_ref = example_base_component.add_data_handler

    target_method_ref(timeseries_axis_name, timeseries_axis.model_copy(deep=True))

    # Assert Axis was add to positions dict
    assert len(example_base_component.positions.items()) == 1

    # Assert correct name was given to Axis within positions dict and that it remained unchanged.
    assert_axis_equality(example_base_component.positions[timeseries_axis_name], timeseries_axis)


@pytest.mark.parametrize("method", ["specific", "generic"])
def test_add_timeseries_position_with_name_conflict(
    example_base_component, example_timeseries_translation_axis, example_timeseries_rotation_axis, method
):
    """
    Validate that add_position and add_data_handler method from Components correctly add Axis with value
    represented as a TimeSeries to its internal positions dictionary and handles name conflicts.

    Expected Behavior:
        - Both added Axis must be within Component positions internal dict
        - The first added Axis must not suffer any change from how it was before
        - Any subsequent Axis added that has the same name with another already existing Axis from the poisitions
        dict from Component must have its name automatically updated to a new unique name by adding an index in
        front the original passed name (e.g: <timeseries_axis_name> (1))

    PS: Current expected behavior should be the same as just adding a normal position with name conflict, as Axis handle
    DataField or TimeSeries objects as its value field.
    """
    timeseries_axis_name = "x"

    if method == "specific":
        target_method_ref = example_base_component.add_position
    if method == "generic":
        target_method_ref = example_base_component.add_data_handler

    target_method_ref(timeseries_axis_name, example_timeseries_translation_axis.model_copy(deep=True))
    target_method_ref(timeseries_axis_name, example_timeseries_rotation_axis.model_copy(deep=True))

    # Assert both Axis objects were add to positions dict
    assert len(example_base_component.positions.items()) == 2

    # 1st Axis name and content must be unchanged
    assert_axis_equality(example_base_component.positions[timeseries_axis_name], example_timeseries_translation_axis)

    # 2nd Axis name must be the same as the 1st with an index due to name conflict and its content must be unchanged
    assert_axis_equality(
        example_base_component.positions[timeseries_axis_name + " (1)"], example_timeseries_rotation_axis
    )


@pytest.mark.parametrize("transformation_type", [t_type for t_type in TransformationType])
@pytest.mark.parametrize("value, unit, metadata, timestamp", create_test_params_combinations(TimeSeries))
def test_add_timeseries_position_methods_consistency(
    value, unit, metadata, timestamp, transformation_type, example_base_component, request
):
    """Validate that both add_position and add_data_handler methods from Components result in
    the same outcome when adding Axis with TimeSeries value to Component.

    Expected Behavior:
        - Both added Axis must be within its respective positions internal dict
        - Both added Axis must be indentified with the same name within its respective positions internal dict
        - Both added Axis must be equivalent
    """
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)
    input_timestamp = get_fixture_values(timestamp, request)
    input_transformation_type = transformation_type

    timeseries_axis = AssonantDataHandlerFactory.create_timeseries_axis(
        transformation_type=input_transformation_type,
        value=input_value,
        timestamps=input_timestamp,
        unit=input_unit,
        extra_metadata=input_metadata,
        timestamps_unit=input_unit,
        timestamp_extra_metadata=input_metadata,
    )
    timeseries_axis_name = "timeseries_axis"

    component_1 = example_base_component.model_copy(deep=True)
    component_2 = example_base_component

    component_1.add_position(timeseries_axis_name, timeseries_axis)
    component_2.add_data_handler(timeseries_axis_name, timeseries_axis)

    assert_axis_equality(component_1.positions[timeseries_axis_name], component_2.positions[timeseries_axis_name])


@pytest.mark.parametrize("transformation_type", [t_type for t_type in TransformationType])
@pytest.mark.parametrize("value, unit, metadata, timestamp", create_test_params_combinations(TimeSeries))
def test_create_and_add_timeseries_position(
    value, unit, metadata, timestamp, transformation_type, example_base_component, request
):
    """
    Validate that create_and_add_timeseries_position method from Components correctly
    add Axis with TimeSeries object to its internal positions dictionary for all accepted data types.

    Expected Behavior:
        - Base Component must create a new Axis with passed data
        - Base Component must add the new created Axis within its positions dictionary
        - Added Axis must be identified with the same name it was given during create_and_add_timeseries_position method
        - Axis object created by the Component must preserve passed data values.
    """
    input_value = get_fixture_values(value, request)
    input_unit = get_fixture_values(unit, request)
    input_metadata = get_fixture_values(metadata, request)
    input_timestamp = get_fixture_values(timestamp, request)
    input_transformation_type = transformation_type

    timeseries_axis_name = "timeseries_axis"
    example_base_component.create_and_add_timeseries_position(
        name=timeseries_axis_name,
        transformation_type=input_transformation_type,
        value=input_value,
        timestamps=input_timestamp,
        unit=input_unit,
        extra_metadata=input_metadata,
        timestamps_unit=input_unit,
        timestamp_extra_metadata=input_metadata,
    )

    # Assert Axis was add to positions dict
    assert len(example_base_component.positions.items()) == 1

    # Assert correct name was given to Axis within positions dict and that it remained unchanged.
    assert_axis_content(
        example_base_component.positions[timeseries_axis_name],
        input_value,
        input_unit,
        input_metadata,
        transformation_type,
        input_timestamp,
        input_unit,
        input_metadata,
    )


def test_create_and_add_timeseries_position_with_name_conflict(
    example_base_component, example_float_1d_array, example_int_1d_array, example_unit, example_metadata
):
    """
    Validate that create_and_add_timeseries_position method from Components correctly
    create and add Axis object to its internal positions dictionary and handles name conflict.

    Expected Behavior:
        - Both Axis must be created
        - Both created Axis must be add to Component positions dictionary
        - The first created and added Axis must have its name and content preserved
        - The second created and added Axis must have its content preserved but name updated
        to a new unique name by adding an index in front of the original anem (e.g: <timeseries_axis_name> (1))
    """
    timeseries_axis_name = "timeseries_axis"
    example_base_component.create_and_add_timeseries_position(
        name=timeseries_axis_name,
        transformation_type=TransformationType.TRANSLATION,
        value=example_float_1d_array,
        timestamps=example_float_1d_array,
        unit=example_unit,
        extra_metadata=example_metadata,
        timestamps_unit=example_unit,
        timestamp_extra_metadata=example_metadata,
    )

    example_base_component.create_and_add_timeseries_position(
        name=timeseries_axis_name,
        transformation_type=TransformationType.ROTATION,
        value=example_int_1d_array,
        timestamps=example_int_1d_array,
        unit=example_unit,
        extra_metadata=example_metadata,
        timestamps_unit=example_unit,
        timestamp_extra_metadata=example_metadata,
    )

    # Assert both created Axis were add to fields dict
    assert len(example_base_component.positions.items()) == 2

    # 1st Axis name and content must be unchanged
    assert_axis_content(
        example_base_component.positions[timeseries_axis_name],
        example_float_1d_array,
        example_unit,
        example_metadata,
        TransformationType.TRANSLATION,
        example_float_1d_array,
        example_unit,
        example_metadata,
    )

    # 2nd TimeSeries name must be the same as the 1st with an index due to name conflict and
    # its content must be unchanged
    assert_axis_content(
        example_base_component.positions[timeseries_axis_name + " (1)"],
        example_int_1d_array,
        example_unit,
        example_metadata,
        TransformationType.ROTATION,
        example_int_1d_array,
        example_unit,
        example_metadata,
    )


def test_add_subcomponent(example_base_component, example_subcomponent):
    """
    Validate that add_subcomponent method from Components correctly
    add subcomponents to its internal dictionary.

    Expected Behavior:
        - Both added subcomponents must be within base_component internal sub_components dict
        - Added sub_components must be exactly equal to how it was before being add
    """
    example_subcomponent_2 = example_subcomponent.model_copy(deep=True)
    example_subcomponent_2.name = "new unique name"

    # Add copies to avoid that internal manipulation changes external data and causing possible False-Positive result.
    example_base_component.add_subcomponent(example_subcomponent.model_copy(deep=True))
    example_base_component.add_subcomponent(example_subcomponent_2.model_copy(deep=True))

    assert len(example_base_component.subcomponents.items()) == 2
    assert_component_equality(example_base_component.subcomponents[example_subcomponent.name], example_subcomponent)
    assert_component_equality(example_base_component.subcomponents[example_subcomponent_2.name], example_subcomponent_2)


def test_add_subcomponent_with_name_conflict(example_base_component, example_subcomponent):
    """
    Validate that add_subcomponent method from Components correctly
    handle cases where subcomponent with conflicting name are add.

    Expected Behavior:
        - Both added subcomponents must be within base_component internal sub_components dict
        - The first added sub_components must not suffer any change from how it was before
        - Any subsequent sub_component added that has the same name with another already existing subcomponent
        must have its name automatically updated to a new unique name by adding an index in front the original
        name (e.g: sub_component (1))
    """
    example_subcomponent_2 = example_subcomponent.model_copy(deep=True)

    example_base_component.add_subcomponent(example_subcomponent.model_copy(deep=True))
    example_base_component.add_subcomponent(example_subcomponent_2.model_copy(deep=True))

    assert len(example_base_component.subcomponents.items()) == 2

    # Assert that 1st subcomponent was totally unchanged
    assert_component_equality(example_base_component.subcomponents[example_subcomponent.name], example_subcomponent)

    with pytest.raises(AssertionError):
        # It is expected the assertion to fail, as subcomponent name should have been renamed,
        # otherwise name conflict handling failed
        assert_component_equality(
            example_base_component.subcomponents[example_subcomponent_2.name + " (1)"], example_subcomponent_2
        )

    # Assert it was only Component names that were different and caused previous AssertionError
    renamed_example_subcomponent_2 = example_subcomponent_2.model_copy(deep=True)
    renamed_example_subcomponent_2.name = example_base_component.subcomponents[
        example_subcomponent_2.name + " (1)"
    ].name
    assert_component_equality(
        example_base_component.subcomponents[example_subcomponent_2.name + " (1)"], renamed_example_subcomponent_2
    )


def test_add_list_of_subcomponents(example_base_component, example_subcomponent):
    """
    Validate that add_subcomponent method from Components correctly
    add list of subcomponents to its internal dictionary.

    Expected Behavior:
        - All added subcomponents must be within base_component internal sub_components dict
        - Added sub_components must be exactly equal to how it was before being add
    """
    example_subcomponent_2 = example_subcomponent.model_copy(deep=True)
    example_subcomponent_3 = example_subcomponent.model_copy(deep=True)
    example_subcomponent_2.name = "2nd subcomponent"
    example_subcomponent_3.name = "3rd subcomponent"

    list_of_subcomponents = [
        example_subcomponent.model_copy(deep=True),
        example_subcomponent_2.model_copy(deep=True),
        example_subcomponent_3.model_copy(deep=True),
    ]

    # Add copies to avoid that internal manipulation changes external data and causing possible False-Positive result.
    example_base_component.add_subcomponent(list_of_subcomponents)

    assert len(example_base_component.subcomponents.items()) == len(list_of_subcomponents)
    assert_component_equality(example_base_component.subcomponents[example_subcomponent.name], example_subcomponent)
    assert_component_equality(example_base_component.subcomponents[example_subcomponent_2.name], example_subcomponent_2)
    assert_component_equality(example_base_component.subcomponents[example_subcomponent_3.name], example_subcomponent_3)


def test_add_list_of_subcomponents_with_name_conflict(example_base_component, example_subcomponent):
    """
    Validate that add_subcomponent method from Components correctly handle cases where subcomponent
    with conflicting name exists inside subcomponents list being.

    Expected Behavior:
        - All added subcomponents must be within base_component internal sub_components dict
        - The first added sub_components must not suffer any change from how it was before
        - Any subsequent sub_component added that has the same name with another already existing subcomponent
        must have its name automatically updated to a new unique name by adding an index in front the original
        name (e.g: sub_component (1))
    """
    example_subcomponent_2 = example_subcomponent.model_copy(deep=True)
    example_subcomponent_3 = example_subcomponent.model_copy(deep=True)

    list_of_subcomponents = [
        example_subcomponent.model_copy(deep=True),
        example_subcomponent_2.model_copy(deep=True),
        example_subcomponent_3.model_copy(deep=True),
    ]

    # Add copies to avoid that internal manipulation changes external data and causing possible False-Positive result.
    example_base_component.add_subcomponent(list_of_subcomponents)

    assert len(example_base_component.subcomponents.items()) == len(list_of_subcomponents)

    # Assert that 1st subcomponent was totally unchanged
    assert_component_equality(example_base_component.subcomponents[example_subcomponent.name], example_subcomponent)

    with pytest.raises(AssertionError):
        # It is expected the assertion to fail, as subcomponent name should have been renamed,
        # otherwise name conflict handling failed
        assert_component_equality(
            example_base_component.subcomponents[example_subcomponent_2.name + " (1)"], example_subcomponent_2
        )

    # Assert it was only Component names that were different and caused previous AssertionError
    renamed_example_subcomponent_2 = example_subcomponent_2.model_copy(deep=True)
    renamed_example_subcomponent_2.name = example_base_component.subcomponents[
        example_subcomponent_2.name + " (1)"
    ].name
    assert_component_equality(
        example_base_component.subcomponents[example_subcomponent_2.name + " (1)"], renamed_example_subcomponent_2
    )

    with pytest.raises(AssertionError):
        # It is expected the assertion to fail, as subcomponent name should have been renamed,
        # otherwise name conflict handling failed
        assert_component_equality(
            example_base_component.subcomponents[example_subcomponent_3.name + " (2)"], example_subcomponent_3
        )

    # Assert it was only Component names that were different and caused previous AssertionError
    renamed_example_subcomponent_3 = example_subcomponent_3.model_copy(deep=True)
    renamed_example_subcomponent_3.name = example_base_component.subcomponents[
        example_subcomponent_3.name + " (2)"
    ].name
    assert_component_equality(
        example_base_component.subcomponents[example_subcomponent_3.name + " (2)"], renamed_example_subcomponent_3
    )


def test_component_take_data_handlers_from_single_component(
    example_base_component,
    example_int_1d_array_data_field,
    example_translation_axis,
    example_float_2d_timeseries_data_field,
    example_timeseries_rotation_axis,
):
    """
    Validate that take_data_handlers_from method from Components correctly take data_handlers
    from a single passed target Component.

    Expected Behavior:
        - All DataHandlers from target Component must be transfered to taker Component
        - Target Component fields and positions internal dicts must be empty after calling the method
        - Taker Component must have in its fields and positions internal dicts the taken DataHandlers
        - DataHandlers must be correctly stored in the correct taker Component internal dicts based in
        its type (Axis -> positions | TimeSeries and DataField -> fields)
        - Taken DataHandlers must be equivalent as they were before being taken
    """

    target_component = example_base_component.model_copy(deep=True)
    taker_component = example_base_component.model_copy(deep=True)

    data_field_name = "data_field"
    timeseries_data_field_name = "timeseries_data_field"
    axis_name = "axis"
    timeseries_axis_name = "timeseries_axis"
    target_component.add_data_handler(data_field_name, example_int_1d_array_data_field.model_copy(deep=True))
    target_component.add_data_handler(axis_name, example_translation_axis.model_copy(deep=True))
    target_component.add_data_handler(
        timeseries_data_field_name, example_float_2d_timeseries_data_field.model_copy(deep=True)
    )
    target_component.add_data_handler(timeseries_axis_name, example_timeseries_rotation_axis.model_copy(deep=True))

    # Check Component status before data handler takage
    assert len(target_component.fields) == 2 and len(target_component.positions) == 2
    assert len(taker_component.fields) == 0 and len(taker_component.positions) == 0

    taker_component.take_data_handlers_from(target_component)

    # Check Component status after data handler takage
    assert len(taker_component.fields) == 2 and len(taker_component.positions) == 2
    assert len(target_component.fields) == 0 and len(target_component.positions) == 0

    # Check if taken fields and positions haven't changed
    assert_data_fields_equality(taker_component.fields[data_field_name], example_int_1d_array_data_field)
    assert_axis_equality(taker_component.positions[axis_name], example_translation_axis)
    assert_timeseries_equality(
        taker_component.fields[timeseries_data_field_name], example_float_2d_timeseries_data_field
    )
    assert_axis_equality(taker_component.positions[timeseries_axis_name], example_timeseries_rotation_axis)


def test_component_take_data_handlers_from_list_of_components(
    example_base_component,
    example_int_1d_array_data_field,
    example_translation_axis,
    example_float_2d_timeseries_data_field,
    example_timeseries_rotation_axis,
):
    """
    Validate that take_data_handlers_from method from Components correctly take data_handlers
    from a list of target Components.

    Expected Behavior:
        - All DataHandlers from target Components must be transfered to taker Component
        - Target Components fields and positions internal dicts must be empty after calling the method
        - Taker Component must have in its fields and positions internal dicts the taken DataHandlers
        from all target Components
        - DataHandlers must be correctly stored in the correct taker Component internal dicts based in
        its type (Axis -> positions | TimeSeries and DataField -> fields)
        - Taken DataHandlers must be equivalent as they were before being taken
    """

    target_component_1 = example_base_component.model_copy(deep=True)
    target_component_2 = example_base_component.model_copy(deep=True)
    target_component_3 = example_base_component.model_copy(deep=True)
    target_component_4 = example_base_component.model_copy(deep=True)
    taker_component = example_base_component.model_copy(deep=True)

    data_field_name = "data_field"
    timeseries_data_field_name = "timeseries_data_field"
    axis_name = "axis"
    timeseries_axis_name = "timeseries_axis"
    target_component_1.add_data_handler(data_field_name, example_int_1d_array_data_field.model_copy(deep=True))
    target_component_2.add_data_handler(axis_name, example_translation_axis.model_copy(deep=True))
    target_component_3.add_data_handler(
        timeseries_data_field_name, example_float_2d_timeseries_data_field.model_copy(deep=True)
    )
    target_component_4.add_data_handler(timeseries_axis_name, example_timeseries_rotation_axis.model_copy(deep=True))

    # Check Components status before data handler takage
    assert len(target_component_1.fields) == 1 and len(target_component_1.positions) == 0
    assert len(target_component_2.fields) == 0 and len(target_component_2.positions) == 1
    assert len(target_component_3.fields) == 1 and len(target_component_3.positions) == 0
    assert len(target_component_4.fields) == 0 and len(target_component_4.positions) == 1
    assert len(taker_component.fields) == 0 and len(taker_component.positions) == 0

    components_list = [target_component_1, target_component_2, target_component_3, target_component_4]

    taker_component.take_data_handlers_from(components_list)

    # Check Components status after data handler takage
    assert len(target_component_1.fields) == 0 and len(target_component_1.positions) == 0
    assert len(target_component_2.fields) == 0 and len(target_component_2.positions) == 0
    assert len(target_component_3.fields) == 0 and len(target_component_3.positions) == 0
    assert len(target_component_4.fields) == 0 and len(target_component_4.positions) == 0
    assert len(taker_component.fields) == 2 and len(taker_component.positions) == 2

    # Check if taken fields and positions haven't changed
    assert_data_fields_equality(taker_component.fields[data_field_name], example_int_1d_array_data_field)
    assert_axis_equality(taker_component.positions[axis_name], example_translation_axis)
    assert_timeseries_equality(
        taker_component.fields[timeseries_data_field_name], example_float_2d_timeseries_data_field
    )
    assert_axis_equality(taker_component.positions[timeseries_axis_name], example_timeseries_rotation_axis)


def test_component_take_data_handlers_with_field_name_conflict(
    example_base_component, example_int_1d_array_data_field, example_float_2d_timeseries_data_field
):
    """
    Validate that take_data_handlers_from method from Components correctly take field data_handlers
    from a Component and handle name conflicts.

    Expected Behavior:
        - All DataHandlers from target Component must be transfered to taker Component even in case
        of name conflict
        - In the case of conflicts, the 1st field added to Component fields dict must be unaltered
        - Additional fields with conflicting names must be renamed by adding an index in front of
        the original name (e.g: <field_name> (1))
    """

    target_component = example_base_component.model_copy(deep=True)
    taker_component = example_base_component.model_copy(deep=True)

    data_field_name = "data_field"

    # Create a situation where field already exists on taker component and will receive a field with the same name
    taker_component.add_data_handler(data_field_name, example_int_1d_array_data_field.model_copy(deep=True))
    target_component.add_data_handler(data_field_name, example_float_2d_timeseries_data_field.model_copy(deep=True))

    taker_component.take_data_handlers_from(target_component)

    # Check if taken fields content haven't changed, except by taken field name
    assert_data_fields_equality(taker_component.fields[data_field_name], example_int_1d_array_data_field)
    assert_timeseries_equality(taker_component.fields[data_field_name + " (1)"], example_float_2d_timeseries_data_field)


def test_component_take_data_handlers_with_position_name_conflict(
    example_base_component, example_translation_axis, example_timeseries_rotation_axis
):
    """
    Validate that take_data_handlers_from method from Components correctly take position data_handlers
    from a Component and handle name conflicts.

    Expected Behavior:
        - All DataHandlers from target Component must be transfered to taker Component even in case
        of name conflict
        - In the case of conflicts, the 1st field added to Component fields dict must be unaltered
        - Additional fields with conflicting names must be renamed by adding an index in front of
        the original name (e.g: <field_name> (1))
    """

    target_component = example_base_component.model_copy(deep=True)
    taker_component = example_base_component.model_copy(deep=True)

    axis_name = "x"

    # Create a situation where position already exists on taker component and will receive a position with the same name
    taker_component.add_data_handler(axis_name, example_translation_axis.model_copy(deep=True))
    target_component.add_data_handler(axis_name, example_timeseries_rotation_axis.model_copy(deep=True))

    taker_component.take_data_handlers_from(target_component)

    # Check if taken fields content haven't changed, except by taken field name
    assert_axis_equality(taker_component.positions[axis_name], example_translation_axis)
    assert_axis_equality(taker_component.positions[axis_name + " (1)"], example_timeseries_rotation_axis)
